import abc

from cliff.lister import Lister as CliffLister
from cliff.show import ShowOne as CliffShowOne
from cliff.command import Command as CliffCommand
from requests import HTTPError
from rich.console import Console

from ..error_handlers import CLIError
from ..utils import tabulate


class ErrorHandlingMixin:
    def run(self, parsed_args):
        try:
            return super(ErrorHandlingMixin, self).run(parsed_args)
        except Exception as exp:
            cli_error = self.on_exception(parsed_args, exp)
            self.represent_error(cli_error)
            return cli_error.return_code

    def on_exception(self, parsed_args, exp):
        try:
            if isinstance(exp, CLIError):
                return exp
            if isinstance(exp, HTTPError):
                result = self._handle_http_error(parsed_args, exp.response)
                if result:
                    return result
            return self.app.error_handler.handle(exp)
        except Exception:
            raise exp

    def represent_error(self, cli_error: CLIError):
        c = self.app.console
        assert isinstance(c, Console)
        c.print("Unable to run your command: ", style="red")
        for error in cli_error.errors:
            c.print("\t* {}".format(error['msg']))
            if 'hint' in error:
                c.print("\t  Hint: {}".format(error['hint']), style='yellow')

    def _handle_http_error(self, parsed_args, response):
        error_handler = getattr(self, "_on_{}".format(response.status_code), None)
        if callable(error_handler):
            return error_handler(parsed_args, response)


class SchemaMixin:
    schema = None

    def get_data(self, parsed_args):
        raise NotImplementedError()

    def take_action(self, parsed_args):
        data = self.get_data(parsed_args)
        if self.schema:
            return tabulate(self.schema.dump(data))
        return tabulate(data)


class Lister(ErrorHandlingMixin, SchemaMixin, CliffLister):

    @property
    def formatter_namespace(self):
        return 'digicloud.formatters.list'

    @property
    def formatter_default(self):
        return "rich"

    @abc.abstractmethod
    def get_data(self, parsed_args):
        raise NotImplementedError()


class ShowOne(ErrorHandlingMixin, SchemaMixin, CliffShowOne):
    @property
    def formatter_namespace(self):
        return 'digicloud.formatters.details'

    @property
    def formatter_default(self):
        return "rich"

    @abc.abstractmethod
    def get_data(self, parsed_args):
        raise NotImplementedError()


class Command(ErrorHandlingMixin, CliffCommand):
    pass



"""
    DigiCloud Compute Instance Service.
"""
import os

from .base import ShowOne, Lister, Command
from .. import schemas
from ..error_handlers import CLIError


class CreateSSHKey(ShowOne):
    """Create new public or private key for server ssh access."""
    schema = schemas.SSHKeyDetails()

    def get_parser(self, prog_name):
        parser = super(CreateSSHKey, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='New public or private key name'
        )
        parser.add_argument(
            '--public-key',
            metavar='<file>',
            required=True,
            help='Filename for public key to add. If not used,'
                 ' creates a private key.'
        )
        return parser

    def get_data(self, parsed_args):
        key_path = os.path.expanduser(parsed_args.public_key)
        with open(key_path) as file_:
            public_key = file_.read()

        payload = {
            'name': parsed_args.name,
            'public_key': public_key or '',
        }

        data = self.app.session.post('/ssh-keys', payload)

        return data

    def on_exception(self, parsed_args, exp):
        if isinstance(exp, FileNotFoundError):
            return CLIError([
                dict(
                    msg="Unable to find your public key file",
                    hint="File doesn't exist, it might be a typo in the file path"
                )
            ])
        if isinstance(exp, PermissionError):
            return CLIError([
                dict(
                    msg="Unable to access your public key file",
                    hint="You public key file has a restrictive permission"
                )
            ])
        return super(CreateSSHKey, self).on_exception(parsed_args, exp)


class ListSSHKey(Lister):
    """List ssh_keys."""
    schema = schemas.SSHKeyList(many=True)

    def get_data(self, parsed_args):
        data = self.app.session.get('/ssh-keys')

        return data


class ShowSSHKey(ShowOne):
    """Show ssh_key detail"""
    schema = schemas.SSHKeyDetails()

    def get_parser(self, prog_name):
        parser = super(ShowSSHKey, self).get_parser(prog_name)
        parser.add_argument(
            'ssh_key',
            metavar='<ssh_key>',
            help=('SSH key Name'),
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/ssh-keys/%s' % parsed_args.ssh_key
        data = self.app.session.get(uri)

        return data


class DeleteSSHKey(Command):
    """Delete SSHKey."""

    def get_parser(self, prog_name):
        parser = super(DeleteSSHKey, self).get_parser(prog_name)
        parser.add_argument(
            'ssh_key',
            metavar='<ssh_key>',
            help=('SSH Key Name')
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/ssh-keys/%s' % parsed_args.ssh_key
        self.app.session.delete(uri)



"""
    Main app components definition.
"""
import sys

from cmd2 import Cmd2ArgparseError
from pkg_resources import get_distribution
from cliff.commandmanager import CommandManager
from sentry_sdk import capture_message

from digicloud.cli.app import BaseApp
from digicloud.error_handlers import ErrorHandler
from .middlewares import (
    ConfigMiddleware,
    SessionMiddleware,
    SignalHandlerMiddleware,
    VersionCheckMiddleware,
    RichMiddleware,
    SentryMiddleware
)


class DigicloudApp(BaseApp):
    def __init__(self):
        command_manager = CommandManager('digicloud.cli')
        self.current_version = get_distribution('digicloud').version
        super(DigicloudApp, self).__init__(
            description="Digicloud CLI",
            version=self.current_version,
            command_manager=command_manager,
            deferred_help=True
        )
        self.middlewares = [
            SignalHandlerMiddleware(self),
            ConfigMiddleware(self),
            SessionMiddleware(self),
            RichMiddleware(self),
            VersionCheckMiddleware(self),
            SentryMiddleware(self),
        ]
        self.error_handler = ErrorHandler(self)

    def clean_up(self, cmd, result, err):
        super(DigicloudApp, self).clean_up(cmd, result, err)
        if isinstance(err, Cmd2ArgparseError):
            capture_message('invalid-command: {}'.format(' '.join(sys.argv[1:])))


def main(argv=None):
    """Initialize main ``cliff.app.App`` instance and run.

    Cliff look for this function as a console script entry point.
    """
    if not argv:
        argv = sys.argv[1:]
    if len(argv) == 0:  # Disable interactive mode
        argv = ['--help']  # display --help instead of interactive mode
    return DigicloudApp().run(argv)


if __name__ == '__main__':
    sys.exit(main())



"""
    DigiCloud load balancer Service.
"""
import ipaddress
from ..utils import is_tty
from .base import Lister, ShowOne, Command
from .. import schemas
from ..error_handlers import CLIError
from rich.prompt import Confirm


################################################
# Load Balancer                                #
################################################
class ListLoadBalancer(Lister):
    """List load balancers"""
    schema = schemas.LoadBalancerList(many=True)

    def get_data(self, parsed_args):
        load_balancer_list = self.app.session.get('/load-balancers')
        return load_balancer_list


class ShowLoadBalancer(ShowOne):
    """Show load balancer details."""
    schema = schemas.LoadBalancerDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/%s' % parsed_args.load_balancer
        load_balancer = self.app.session.get(uri)
        return load_balancer


class DeleteLoadBalancer(Command):
    """Delete load balancer."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )

        return parser

    def take_action(self, parsed_args):
        self.confirm_deletion(parsed_args)
        uri = '/load-balancers/%s' % parsed_args.load_balancer
        self.app.session.delete(uri)

    def confirm_deletion(self, parsed_args):
        if parsed_args.i_am_sure:
            return True
        if is_tty():
            user_response = Confirm.ask(
                "You're about to delete loadbalancer [red bold]{}[/red bold]. "
                "Are you sure?".format(
                    parsed_args.load_balancer
                ), default=False
            )
            if user_response:
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write(
                "Unable to perform 'loadbalancer delete' operation in non-interactive mode,"
                " without '--i-am-sure' switch\n")
            return False


class UpdateLoadBalancer(ShowOne):
    """Update load balancer."""
    schema = schemas.LoadBalancerDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--name',
            metavar='<Name>',
            help='New name for load balancer.'
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help='Load balancer description.',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/%s' % parsed_args.load_balancer
        payload = {}
        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.description:
            payload['description'] = parsed_args.description

        if not payload:
            raise CLIError([dict(
                msg="At least one of "
                    " --name or"
                    " --description is necessary"
            )])
        load_balancer = self.app.session.patch(uri, payload)
        return load_balancer


class CreateLoadBalancer(ShowOne):
    """Create Load Balancer"""
    schema = schemas.LoadBalancerDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<Name>',
            help='New name for load balancer.',
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help='Load balancer description.',
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            'name': parsed_args.name,
        }
        if parsed_args.description:
            payload['description'] = parsed_args.description
        load_balancer = self.app.session.post('/load-balancers', payload)
        return load_balancer


################################################
# Backend                                      #
################################################


class CreateBackend(ShowOne):
    """Create Load balancer Backend.

    Each backend may contain multiple backend members.
    """
    schema = schemas.BackendDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--algorithm',
            metavar='<algorithm>',
            help='Backend algorithm one of "round_robin" or "source_ip"',
            choices=['round_robin', 'source_ip'],
            required=True
        )
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            'algorithm': parsed_args.algorithm,
        }
        uri = '/load-balancers/%s/backends' % parsed_args.load_balancer
        backend = self.app.session.post(uri, payload)
        return backend


class DeleteBackend(Command):
    """Delete a LoadBalancer backend."""
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--backend-id',
            metavar='<backend_id>',
            required=True,
            help='The backend identifier to be deleted.',
        )
        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )

        return parser

    def take_action(self, parsed_args):
        self.confirm_deletion(parsed_args)
        uri = '/load-balancers/%s/backends/%s' % (
            parsed_args.load_balancer,
            parsed_args.backend_id,
        )
        self.app.session.delete(uri)

    def confirm_deletion(self, parsed_args):
        if parsed_args.i_am_sure:
            return True
        if is_tty():
            user_response = Confirm.ask(
                "You're about to delete backend "
                " [red bold]{}[/red bold] "
                "including the members and the health check. "
                "Are you sure?".format(
                    parsed_args.backend_id
                ), default=False
            )
            if user_response:
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write(
                "Unable to perform 'backend delete' operation in non-interactive mode,"
                " without '--i-am-sure' switch\n")
            return False


class ListBackend(Lister):
    """Lists a load balancer backends."""
    schema = schemas.BackendDetail(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/%s/backends' % parsed_args.load_balancer
        backends = self.app.session.get(uri)
        return backends


################################################
# Frontend                                     #
################################################


class CreateFrontend(ShowOne):
    """Create Load balancer Frontend."""
    schema = schemas.FrontendDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='The TCP port which frontend listens to.',
            type=int,
            required=True,
        )
        parser.add_argument(
            '--timeout-client-data',
            metavar='<timeout_client_data>',
            help='Client inactivity timeout in milliseconds.',
            type=int,
            required=False,
        )
        parser.add_argument(
            '--timeout-tcp-inspect',
            metavar='<timeout_tcp_inspect>',
            help='Time, in milliseconds, '
                 'to wait for additional TCP packets for content inspection.',
            type=int,
            required=False,
        )
        parser.add_argument(
            '--timeout-member-connect',
            metavar='<timeout_member_connect>',
            help='Backend member connection timeout in milliseconds.',
            type=int,
            required=False,
        )
        parser.add_argument(
            '--timeout_member_data',
            metavar='<timeout_member_data>',
            help='Backend member inactivity timeout in milliseconds.',
            type=int,
            required=False,
        )
        parser.add_argument(
            '--connection-limit',
            metavar='<connection_limit>',
            help='The maximum number of connections permitted for this Frontend.'
                 'Value "-1" represents infinite connections.',
            type=int,
            required=True,
        )
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--default-backend',
            metavar='<default_backend>',
            required=False,
            help='default backend which the frontend will forward requests to.',
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            'port': parsed_args.port,
            'default_backend_id': parsed_args.default_backend,
            "connection_limit": parsed_args.connection_limit,
        }
        if parsed_args.timeout_client_data:
            payload["timeout_client_data"] = parsed_args.timeout_client_data
        if parsed_args.timeout_tcp_inspect:
            payload["timeout_tcp_inspect"] = parsed_args.timeout_tcp_inspect
        if parsed_args.timeout_member_connect:
            payload["timeout_member_connect"] = parsed_args.timeout_member_connect
        if parsed_args.timeout_member_data:
            payload["timeout_member_data"] = parsed_args.timeout_member_data

        uri = '/load-balancers/%s/frontends' % parsed_args.load_balancer
        frontend = self.app.session.post(uri, payload)
        return frontend


class SetFrontendDefaultBackend(ShowOne):
    """Set a new default backend for the specified frontend."""
    schema = schemas.FrontendDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--frontend-id',
            metavar='<frontend_id>',
            required=True,
            help='The frontend identifier to be updated.',
        )
        parser.add_argument(
            '--default-backend',
            metavar='<default_backend>',
            required=True,
            help='default backend which the frontend will forward requests to.',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/{}/frontends/{}'.format(
            parsed_args.load_balancer, parsed_args.frontend_id
        )
        payload = {"default_backend_id": parsed_args.default_backend}
        frontend = self.app.session.patch(uri, payload)
        return frontend


class DeleteFrontendDefaultBackend(ShowOne):
    """Delete default backend for the specified frontend."""
    schema = schemas.FrontendDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--frontend-id',
            metavar='<frontend_id>',
            required=True,
            help='The frontend identifier to be updated.',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/{}/frontends/{}'.format(
            parsed_args.load_balancer, parsed_args.frontend_id
        )
        payload = {"default_backend_id": None}
        frontend = self.app.session.patch(uri, payload)
        return frontend


class DeleteFrontend(Command):
    """Delete a LoadBalancer frontend."""
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--frontend-id',
            metavar='<frontend_id>',
            required=True,
            help='The frontend identifier to be deleted.',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/load-balancers/%s/frontends/%s' % (
            parsed_args.load_balancer,
            parsed_args.frontend_id,
        )
        self.app.session.delete(uri)


class ListFrontend(Lister):
    """Lists a load balancer frontends."""
    schema = schemas.FrontendDetail(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/%s/frontends' % parsed_args.load_balancer
        frontends = self.app.session.get(uri)
        return frontends

################################################
# Health Check                                 #
################################################


class ListHealthCheck(Lister):
    """Lists backend Health check."""
    schema = schemas.HealthCheckDetail(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--backend-id',
            metavar='<backend_id>',
            required=True,
            help='The backend identifier.',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/%s/backends/%s/health-checks' % (
            parsed_args.load_balancer,
            parsed_args.backend_id,
        )
        health_checks = self.app.session.get(uri)
        return health_checks


class DeleteHealthCheck(Command):
    """Delete Health Check."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--backend-id',
            metavar='<backend_id>',
            required=True,
            help='The backend identifier.',
        )
        parser.add_argument(
            '--health-check-id',
            metavar='<health_check_id>',
            required=True,
            help='The health check identifier to be deleted.',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/load-balancers/%s/backends/%s/health-checks/%s' % (
            parsed_args.load_balancer,
            parsed_args.backend_id,
            parsed_args.health_check_id,
        )
        self.app.session.delete(uri)


class CreateHealthCheck(ShowOne):
    """Create Health Check"""
    schema = schemas.HealthCheckDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--backend-id',
            metavar='<backend_id>',
            required=True,
            help='The backend identifier.',
        )
        parser.add_argument(
            '--name',
            metavar='<Name>',
            help='Human readable name.',
            required=True,
        )
        parser.add_argument(
            '--delay',
            metavar='<delay>',
            help='Health Check delay.',
            type=int,
            required=True,
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            help='Health Check timeout.',
            type=int,
            required=True
        )
        parser.add_argument(
            '--max-retries',
            metavar='<max-retries>',
            help='Health Check max retries.',
            type=int,
            required=True,
        )
        parser.add_argument(
            '--max-retries-down',
            metavar='<max-retries-down>',
            help='The number of allowed check failures before '
                 'changing the operating status of the member to ERROR. '
                 'A valid value is from 1 to 10.',
            type=int,
            required=True,
        )
        return parser

    def _check_arg_validity(self, parsed_args):
        rules = [
            (
                parsed_args.delay < parsed_args.timeout,
                "timeout must be lower than and equal to delay",
            ),
        ]
        errors = []
        for is_invalid, err_msg in rules:
            if is_invalid:
                errors.append(dict(msg=err_msg))
        if errors:
            raise CLIError(errors)

    def get_data(self, parsed_args):
        self._check_arg_validity(parsed_args)
        uri = '/load-balancers/%s/backends/%s/health-checks' % (
            parsed_args.load_balancer,
            parsed_args.backend_id,
        )
        payload = {
            'name': parsed_args.name,
            'delay': parsed_args.delay,
            'timeout': parsed_args.timeout,
            'max_retries': parsed_args.max_retries,
            'max_retries_down': parsed_args.max_retries_down,
        }
        health_check = self.app.session.post(uri, payload)
        return health_check


################################################
# Backend Member                               #
################################################

class ListBackendMember(Lister):
    """Lists backend members."""
    schema = schemas.BackendMemberDetail(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--backend-id',
            metavar='<backend_id>',
            required=True,
            help='The backend identifier.',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/%s/backends/%s/members' % (
            parsed_args.load_balancer,
            parsed_args.backend_id,
        )
        members = self.app.session.get(uri)
        return members


class DeleteBackendMember(Command):
    """Delete backend member."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--backend-id',
            metavar='<backend_id>',
            required=True,
            help='The backend identifier.',
        )
        parser.add_argument(
            '--member-id',
            metavar='<member_id>',
            help='The member identifier to be deleted.',
            required=True
        )
        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )
        return parser

    def take_action(self, parsed_args):
        self.confirm_deletion(parsed_args)
        uri = '/load-balancers/%s/backends/%s/members/%s' % (
            parsed_args.load_balancer,
            parsed_args.backend_id,
            parsed_args.member_id,
        )
        self.app.session.delete(uri)

    def confirm_deletion(self, parsed_args):
        if parsed_args.i_am_sure:
            return True
        if is_tty():
            user_response = Confirm.ask(
                "You're about to delete backend member"
                " [red bold]{}[/red bold] . "
                "Are you sure?".format(
                    parsed_args.member_id
                ), default=False
            )
            if user_response:
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write(
                "Unable to perform 'backend member delete' operation in non-interactive mode,"
                " without '--i-am-sure' switch\n")
            return False


class AddBackendMember(Lister):
    """Add Backend member"""
    schema = schemas.BackendMemberDetail(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--backend-id',
            metavar='<backend_id>',
            required=True,
            help='The backend identifier.',
        )
        parser.add_argument(
            '--ip-address',
            metavar='<ip-address>',
            help='Backend member ip address.',
            required=True,
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Backend member port.',
            type=int,
            required=True,
        )
        return parser

    def _check_ip_address(self, value):
        try:
            ipaddress.IPv4Address(value)
            return False
        except (ipaddress.AddressValueError, ipaddress.NetmaskValueError):
            return True

    def _check_arg_validity(self, parsed_args):
        rules = [
            (
                parsed_args.port and (parsed_args.port < 1 or parsed_args.port > 65535),
                "port should be in range 1 - 65535",
            ),
            (
                self._check_ip_address(parsed_args.ip_address),
                "Not a valid ipv4 address"
            )
        ]
        errors = []
        for is_invalid, err_msg in rules:
            if is_invalid:
                errors.append(dict(msg=err_msg))
        if errors:
            raise CLIError(errors)

    def get_data(self, parsed_args):
        self._check_arg_validity(parsed_args)
        uri = '/load-balancers/%s/backends/%s/members' % (
            parsed_args.load_balancer,
            parsed_args.backend_id,
        )
        payload = [{
            'ip_address': parsed_args.ip_address,
            'port': parsed_args.port,
        }]
        backend_member = self.app.session.post(uri, payload)
        return backend_member



from rich.prompt import Confirm

from digicloud import schemas
from .base import Lister, ShowOne, Command
from ..utils import is_tty


class CreateSnapshot(ShowOne):
    """Create snapshot."""
    schema = schemas.SnapshotDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Snapshot name',
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help='Snapshot description',
            required=False,
        )
        parser.add_argument(
            '--instance',
            required=True,
            metavar='<instance>',
            help='Instance name or ID'
        )

        return parser

    def get_data(self, parsed_args):
        self.app.console.print(
            "[yellow bold]"
            "WARNING: If your instance has any volume larger than 100G,"
            " you won't be able to create an instance from this snapshot."
            " Please contact support for help.\n"
            "[/yellow bold]"
        )
        uri = '/snapshots'
        payload = {
            'name': parsed_args.name,
            'instance': parsed_args.instance,
        }
        if parsed_args.description:
            payload['description'] = parsed_args.description
        data = self.app.session.post(uri, payload)
        return data


class ListSnapshot(Lister):
    """List snapshots."""
    schema = schemas.SnapshotList(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        return parser

    def get_data(self, parsed_args):
        uri = '/snapshots'
        data = self.app.session.get(uri)
        return data


class ShowSnapshot(ShowOne):
    """Show snapshot details."""
    schema = schemas.SnapshotDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help='Snapshot name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/snapshots/%s' % parsed_args.snapshot
        data = self.app.session.get(uri)
        return data


class DeleteSnapshot(Command):
    """Delete snapshot."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help='snapshot name or ID'
        )
        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )
        return parser

    def take_action(self, parsed_args):
        if not self.confirm(parsed_args):
            return
        uri = '/snapshots/%s' % parsed_args.snapshot
        self.app.session.delete(uri)

    def confirm(self, parsed_args):
        if parsed_args.i_am_sure:
            return True
        if is_tty():
            snapshot = self.app.session.get(
                '/snapshots/%s' % parsed_args.snapshot
            )
            user_response = Confirm.ask(
                "You're about to delete a snapshot named "
                "[red bold]{}[/red bold]. "
                "Are you sure?".format(
                    snapshot['name'],
                ))
            if user_response:
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write(
                "Unable to perform 'delete snapshot' operation in "
                "non-interactive mode, without '--i-am-sure' switch\n"
            )
            return False


class UpdateSnapshot(ShowOne):
    """Update snapshot name or description."""
    schema = schemas.SnapshotDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help='Snapshot name or ID',
        )
        parser.add_argument(
            '--description',
            metavar='<new_description>',
            help='Snapshot new description',
            required=False,
        )
        parser.add_argument(
            '--name',
            metavar='<new_name>',
            help='Snapshot new name',
            required=False,
        )
        return parser

    def get_data(self, parsed_args):
        uri = f'/snapshots/{parsed_args.snapshot}'

        payload = {}
        if parsed_args.name:
            payload["name"] = parsed_args.name

        if parsed_args.description:
            payload["description"] = parsed_args.description

        data = self.app.session.patch(uri, payload)
        return data



"""
    DigiCloud Edge SND and CDN services.
"""
import json
import os

from marshmallow.exceptions import ValidationError
from rich.prompt import Confirm, Prompt

from .base import Lister, ShowOne, Command
from ..error_handlers import CLIError
from .. import schemas
from ..utils import is_tty, t_or_f_or_n, tabulate as util_tabulate
from tabulate import tabulate


class ListDomain(Lister):
    """List edge domains"""
    schema = schemas.EdgeDomainList(many=True)

    def get_data(self, parsed_args):
        domains = self.app.session.get('/edge/domains')
        return domains


class CreateDomain(ShowOne):
    """Create Domain"""
    schema = schemas.EdgeDomainDetails()
    package_schema = schemas.EdgeDomainPackageDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Domain name',
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            'name': parsed_args.name,
            'package_id': self._get_package(),
        }
        domain = self.app.session.post('/edge/domains', payload)

        if domain['is_subdomain']:
            self.app.console.print(
                "[bold yellow]{}[/bold yellow]"
                "In order to define a sub-domain in DIGICLOUD, "
                "you need to verify your ownership on Root Domain. "
                "Please add a TXT record on your root Domain DNS Setting".format(
                    "Create a TXT Record like the following: "
                    "\n\t [bold blue]_digicdn_check.{}[/bold blue] = "
                    "[bold blue]{}[/bold blue] \n".format(
                        str(parsed_args.name).split(".")[0], domain["ns_verification"]
                    )
                )
            )
        else:
            result = self.app.session.get(
                '/edge/domains/{}/ns-records'.format(domain["id"])
            )
            self.app.console.print(
                "[bold yellow]{}[/bold yellow]".format(
                    "Make sure to set DigiCloud NS records: "
                    "\n\t [bold blue]{}[/bold blue]".format(
                        "\n\t ".join(result["digicloud_ns_records"])
                    )
                )
            )
        return domain

    def _get_package(self):
        result = util_tabulate(
            self.package_schema.dump(
                self.app.session.get(
                    '/edge/packages'
                ),
                many=True,
            )
        )

        self.app.console.print(
            tabulate(
                result[1],
                headers=result[0],
                tablefmt='outline',
            )
        )

        self.app.console.print(
            "[bold yellow]{}[/bold yellow]".format(
                "Please select one from above list for your domain package!"
            )
        )

        return Prompt.ask("Package id")


class GetDomainDNSSEC(ShowOne):
    """Get Domain DNSSEC"""
    schema = schemas.DomainDNSSECDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='Domain name or id',
        )
        return parser

    def get_data(self, parsed_args):
        dnssec = self.app.session.get(
            '/edge/domains/{}/dnssec'.format(parsed_args.domain)
        )
        return dnssec


class EnableDomainDNSSEC(ShowOne):
    """Enable domain dnssec."""
    schema = schemas.DomainDNSSECDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='Domain name or id',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/dnssec'.format(parsed_args.domain)
        payload = {
            "dnssec": True,
        }
        dnssec = self.app.session.patch(uri, payload)
        return dnssec


class DisableDomainDNSSEC(ShowOne):
    """Disable domain dnssec."""
    schema = schemas.DomainDNSSECDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='Domain name or id',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/dnssec'.format(parsed_args.domain)
        payload = {
            "dnssec": False,
        }
        dnssec = self.app.session.patch(uri, payload)
        return dnssec


class GetDomainNSRecordSetting(Command):
    """Get Domain/Sub Domain NS record setting"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='Domain name or id',
        )
        return parser

    def take_action(self, parsed_args):
        domain = self.app.session.get('/edge/domains/%s' % parsed_args.domain)
        if domain['is_subdomain']:
            self.app.console.print(
                "[bold yellow]{}[/bold yellow]"
                "In order to define a sub-domain in DIGICLOUD, "
                "you need to verify your ownership on Root Domain. "
                "Please add a TXT record on your root Domain DNS Setting".format(
                    "Create a TXT Record like the following: "
                    "\n\t [bold blue]_digicdn_check.{}[/bold blue] = "
                    "[bold blue]{}[/bold blue] \n".format(
                        str(domain['name']).split(".")[0], domain["ns_verification"]
                    )
                )
            )
        else:
            result = self.app.session.get(
                '/edge/domains/{}/ns-records'.format(parsed_args.domain)
            )
            self.app.console.print(
                "[bold yellow]{}[/bold yellow]".format(
                    "Please set these NS records: \n"
                    "\t [bold blue]{}[/bold blue]".format(
                        "\n\t ".join(result["digicloud_ns_records"])
                    )
                )
            )


class VerifyNSRecord(Command):
    """Verify ns record"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='Domain name or id',
        )
        return parser

    def take_action(self, parsed_args):
        result = self.app.session.get(
            '/edge/domains/{}/verify-ns-records'.format(parsed_args.domain)
        )
        if result["result"]:
            self.app.console.print(
                "[bold green]{}[/bold green]".format(
                    "DigiCloud NS records fully activated."
                )
            )
        else:
            domain = self.app.session.get('/edge/domains/%s' % parsed_args.domain)
            if domain['is_subdomain']:
                self.app.console.print(
                    "[bold red]{}[/bold red]".format(
                        "DigiCloud was not able to verify your TXT records."
                    )
                )
                self.app.console.print(
                    "[bold yellow]{}[/bold yellow]"
                    "In order to define a sub-domain in DIGICLOUD, "
                    "you need to verify your ownership on Root Domain. "
                    "Please add a TXT record on your root Domain DNS Setting".format(
                        "Create a TXT Record like the following: "
                        "\n\t [bold blue]_digicdn_check.{}[/bold blue] = "
                        "[bold blue]{}[/bold blue] \n".format(
                            str(domain['name']).split(".")[0], domain["ns_verification"]
                        )
                    )
                )
            else:
                result = self.app.session.get(
                    '/edge/domains/{}/ns-records'.format(parsed_args.domain)
                )
                self.app.console.print(
                    "[bold red]{}[/bold red]".format(
                        "DigiCloud was not able to verify your NS records."
                    )
                )
                self.app.console.print(
                    "[bold yellow]{}[/bold yellow]".format(
                        "Make sure to set DigiCloud NS records: \n\t [bold blue]{}[/bold blue]".format(
                            "\n\t ".join(result["digicloud_ns_records"])
                        )
                    )
                )


class DeleteDomain(Command):
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='Domain name or id',
        )

        parser.add_argument(
            '--yes-i-really-mean-it',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )
        return parser

    def take_action(self, parsed_args):
        domain = self.app.session.get('/edge/domains/%s' % parsed_args.domain)
        if not self.delete_domain_2_step(domain):
            return
        if not self.confirm_domain_deletion(parsed_args, domain):
            return
        self.app.session.delete('/edge/domains/%s' % parsed_args.domain)

    def confirm_domain_deletion(self, parsed_args, domain):
        if parsed_args.yes_i_really_mean_it:
            return True
        if is_tty():
            user_response = Confirm.ask(
                "You're about to delete domain named [red bold]{}[/red bold]. "
                "Are you sure?".format(
                    domain['name']
                ), default=False
            )
            if user_response:
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write(
                "Unable to perform 'domain delete' operation in non-interactive mode,"
                " without '--yes-i-really-mean-it' switch\n")
            return False

    def delete_domain_2_step(self, domain):
        if is_tty():
            domain_name = Prompt.ask(
                "[red bold]Attention:[/red bold] "
                "You're about to delete domain named [red bold]{}[/red bold]. "
                "Are you sure? Please write the domain name".format(
                    domain['name'],
                )
            )
            if domain_name == domain["name"]:
                return True
            self.app.stdout.write("Wrong domain name!\n")
            return False
        else:
            return True


class ShowDomain(ShowOne):
    schema = schemas.EdgeDomainDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='Domain name or id',
        )
        return parser

    def get_data(self, parsed_args):
        domain = self.app.session.get('/edge/domains/%s' % parsed_args.domain)
        return domain


def get_record_schema(data):
    schema_map = {
        "A": schemas.ARecordDetailsSchema(),
        "TXT": schemas.TXTRecordDetailsSchema(),
        "CNAME": schemas.CNAMERecordDetailsSchema(),
        "MX": schemas.MXRecordDetailsSchema(),
        "SRV": schemas.SRVRecordDetailsSchema(),
    }
    return schema_map.get(data["type"], None)


class ListRecord(Lister):
    """List edge records"""
    schema = schemas.RecordListSchema(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        domains = self.app.session.get('/edge/domains/{}/records'.format(parsed_args.domain))
        return domains


class ShowRecord(ShowOne):
    """Show domain details."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'record',
            metavar='<record>',
            help='edge record ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/records/{}'.format(parsed_args.domain, parsed_args.record)
        record = self.app.session.get(uri)
        self.schema = get_record_schema(record)
        return record


class DeleteRecord(Command):
    """Delete record."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'record',
            metavar='<record>',
            help='edge record ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/edge/domains/{}/records/{}'.format(parsed_args.domain, parsed_args.record)
        self.app.session.delete(uri)


class UpdateRecord(ShowOne):
    """Update record."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'record',
            metavar='<record>',
            help='edge record ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='Record name.',
            required=False,
        )
        parser.add_argument(
            '--ttl',
            metavar='<ttl>',
            help='Time to live.',
            choices=("2m", "10m", "30m", "1h", "3h", "10h",),
            required=False,
        )
        parser.add_argument(
            '--note',
            metavar='<note>',
            help='Note',
            required=False,
        )
        parser.add_argument(
            '--ip-address',
            metavar='<ip-address>',
            help='IP Address.',
            required=False,
        )

        parser.add_argument(
            '--content',
            metavar='<content>',
            help='Content.',
            required=False,
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Port.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--weight',
            metavar='<weight>',
            help='Weight.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--proto',
            metavar='<proto>',
            help='Proto.',
            required=False,
            choices=("_tcp", "_udp", "_tls"),
        )
        parser.add_argument(
            '--service',
            metavar='<service>',
            help='Service.',
            required=False,
        )
        parser.add_argument(
            '--target',
            metavar='<target>',
            help='Target.',
            required=False,
        )
        parser.add_argument(
            '--mail-server',
            metavar='<mail_server>',
            help='Mail server.',
            required=False,
        )
        parser.add_argument(
            '--priority',
            metavar='<priority>',
            help='Priority.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--proxy',
            metavar='<proxy>',
            help='proxy.',
            required=False,
            type=t_or_f_or_n,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/records/{}'.format(parsed_args.domain, parsed_args.record)
        payload = {
            key: value for key, value in vars(parsed_args).items()
            if value is not None and key in schemas.RecordListSchema.available_keys
        }
        record = self.app.session.patch(uri, payload)
        self.schema = get_record_schema(record)
        return record


class CreateRecord(ShowOne):
    """Create Record"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='Record name.',
            required=True,
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            help='Record type.',
            choices=("A", "TXT", "CNAME", "MX", "SRV",),
            required=True,
            type=str.upper,
        )
        parser.add_argument(
            '--ttl',
            metavar='<ttl>',
            help='Time to live.',
            choices=("2m", "10m", "30m", "1h", "3h", "10h",),
            required=True,
        )
        parser.add_argument(
            '--note',
            metavar='<note>',
            help='Note',
            required=False,
        )
        parser.add_argument(
            '--ip-address',
            metavar='<ip-address>',
            help='IP Address.',
            required=False,
        )

        parser.add_argument(
            '--content',
            metavar='<content>',
            help='Content.',
            required=False,
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Port.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--weight',
            metavar='<weight>',
            help='Weight.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--proto',
            metavar='<proto>',
            help='Proto.',
            required=False,
            choices=("_tcp", "_udp", "_tls"),
        )
        parser.add_argument(
            '--service',
            metavar='<service>',
            help='Service.',
            required=False,
        )
        parser.add_argument(
            '--target',
            metavar='<target>',
            help='Target.',
            required=False,
        )
        parser.add_argument(
            '--mail-server',
            metavar='<mail_server>',
            help='Mail server.',
            required=False,
        )
        parser.add_argument(
            '--priority',
            metavar='<priority>',
            help='Priority.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--proxy',
            metavar='<proxy>',
            help='proxy.',
            required=False,
            type=t_or_f_or_n,
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            key: value for key, value in vars(parsed_args).items()
            if value is not None and key in schemas.RecordListSchema.available_keys
        }
        try:
            payload = self._get_record_type_schema(payload["type"])().load(payload)
        except ValidationError as e:
            raise CLIError(self._handle_validation_error(payload["type"], e))
        record = self.app.session.post('/edge/domains/{}/records'.format(parsed_args.domain), payload)
        self.schema = get_record_schema(record)
        return record

    @staticmethod
    def _handle_validation_error(record_type: str, e: ValidationError):
        errors = []
        for key, value in e.messages.items():
            error = "".join(value)
            if error == "Unknown field.":
                msg = "can not use --{} with {} record type.".format(key, record_type)
            elif error == "Missing data for required field.":
                msg = "--{} is required by {} record type.".format(
                    key.replace("_", "-"),
                    record_type
                )
            else:
                raise NotImplementedError
            errors.append(dict(
                msg=msg
            ))
        return errors

    @staticmethod
    def _get_record_type_schema(record_type: str):
        record_schema_map = {
            "A": schemas.ARecordDetailsSchema,
            "TXT": schemas.TXTRecordDetailsSchema,
            "MX": schemas.MXRecordDetailsSchema,
            "CNAME": schemas.CNAMERecordDetailsSchema,
            "SRV": schemas.SRVRecordDetailsSchema,
        }
        return record_schema_map[record_type]


class ListUpstream(Lister):
    """List edge upstreams"""
    schema = schemas.UpstreamDetails(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        upstreams = self.app.session.get('/edge/domains/{}/upstreams'.format(parsed_args.domain))
        return upstreams


class ShowUpstream(ShowOne):
    """Show upstream details."""
    schema = schemas.UpstreamDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'upstream',
            metavar='<upstream>',
            help='edge upstream name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/upstreams/{}'.format(parsed_args.domain, parsed_args.upstream)
        upstream = self.app.session.get(uri)
        return upstream


class DeleteUpstream(Command):
    """Delete upstream."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'upstream',
            metavar='<upstream>',
            help='edge upstream name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/edge/domains/{}/upstreams/{}'.format(parsed_args.domain, parsed_args.upstream)
        self.app.session.delete(uri)


class UpdateUpstream(ShowOne):
    """Update upstream."""
    schema = schemas.UpstreamDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'upstream',
            metavar='<upstream>',
            help='edge upstream name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='upstream name.',
        )
        parser.add_argument(
            '--lb-method',
            metavar='<lb-method>',
            help='Load balancing method.',
            choices=("consistent_ip_hash", "round_robin",),
            type=str.lower,
        )
        parser.add_argument(
            '--keep-alive',
            metavar='<keep-alive>',
            help='keep alive time.',
        )
        parser.add_argument(
            '--ssl-policy',
            metavar='<ssl-policy>',
            help='SSL policy.',
            choices=("http", "https",),
            type=str.lower,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/upstreams/{}'.format(parsed_args.domain, parsed_args.upstream)
        payload = {}
        if parsed_args.name:
            payload["name"] = parsed_args.name
        if parsed_args.lb_method:
            payload["lb_method"] = parsed_args.lb_method
        if parsed_args.keep_alive:
            payload["keep_alive"] = parsed_args.keep_alive
        if parsed_args.ssl_policy:
            payload["ssl_policy"] = parsed_args.ssl_policy
        upstream = self.app.session.patch(uri, payload)
        return upstream


class CreateUpstream(ShowOne):
    """Create upstream"""
    schema = schemas.UpstreamDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='upstream name.',
            required=True,
        )
        parser.add_argument(
            '--lb-method',
            metavar='<lb-method>',
            help='Load balancing method.',
            choices=("consistent_ip_hash", "round_robin",),
            required=True,
            type=str.lower,
        )
        parser.add_argument(
            '--keep-alive',
            metavar='<keep-alive>',
            help='keep alive time.',
            required=True,
        )
        parser.add_argument(
            '--ssl-policy',
            metavar='<ssl-policy>',
            help='SSL policy.',
            choices=("http", "https",),
            required=True,
            type=str.lower,
        )
        return parser

    def get_data(self, parsed_args):
        url = '/edge/domains/{}/upstreams'.format(parsed_args.domain)
        payload = {
            "name": parsed_args.name,
            "lb_method": parsed_args.lb_method,
            "keep_alive": parsed_args.keep_alive,
            "ssl_policy": parsed_args.ssl_policy,
        }
        upstream = self.app.session.post(url, payload)
        return upstream


class ListUpstreamServer(Lister):
    """List edge upstream servers"""
    schema = schemas.UpstreamServerDetails(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--upstream',
            metavar='<upstream>',
            help='edge upstream name or ID',
            required=True,
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        servers = self.app.session.get(
            '/edge/domains/{}/upstreams/{}/servers'.format(
                parsed_args.domain,
                parsed_args.upstream,
            ))
        return servers


class ShowUpstreamServer(ShowOne):
    """Show upstream server details."""
    schema = schemas.UpstreamServerDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help='edge upstream server ID',
        )
        parser.add_argument(
            '--upstream',
            metavar='<upstream>',
            help='edge upstream name or ID',
            required=True,
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/upstreams/{}/servers/{}'.format(
            parsed_args.domain, parsed_args.upstream, parsed_args.server
        )
        server = self.app.session.get(uri)
        return server


class DeleteUpstreamServer(Command):
    """Delete upstream server."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help='edge upstream server ID',
        )
        parser.add_argument(
            '--upstream',
            metavar='<upstream>',
            help='edge upstream name or ID',
            required=True,
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/edge/domains/{}/upstreams/{}/servers/{}'.format(
            parsed_args.domain, parsed_args.upstream, parsed_args.server
        )
        self.app.session.delete(uri)


class UpdateUpstreamServer(ShowOne):
    """Update upstream server."""
    schema = schemas.UpstreamServerDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help='edge upstream server ID',
        )
        parser.add_argument(
            '--upstream',
            metavar='<upstream>',
            help='edge upstream name or ID',
            required=True,
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        parser.add_argument(
            '--ip-domain',
            metavar='<ip-domain>',
            help='Server ip domain.',
            required=True,
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Server port.',
            required=True,
            type=int,
        )
        parser.add_argument(
            '--weight',
            metavar='<weight>',
            help='server weight.',
            required=True,
            type=int,
        )
        parser.add_argument(
            '--fail-timeout',
            metavar='<fail-timeout>',
            help='Server fail timeout.',
            required=True,
            type=int,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/upstreams/{}/servers/{}'.format(
            parsed_args.domain, parsed_args.upstream, parsed_args.server
        )
        payload = {}
        if parsed_args.ip_domain:
            payload["ip_domain"] = parsed_args.ip_domain
        if parsed_args.port:
            payload["port"] = parsed_args.port
        if parsed_args.weight:
            payload["weight"] = parsed_args.weight
        if parsed_args.fail_timeout:
            payload["fail_timeout"] = parsed_args.fail_timeout
        server = self.app.session.patch(uri, payload)
        return server


class CreateUpstreamServer(ShowOne):
    """Create upstream server"""
    schema = schemas.UpstreamServerDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--upstream',
            metavar='<upstream>',
            help='edge upstream name or ID',
            required=True,
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        parser.add_argument(
            '--ip-domain',
            metavar='<ip-domain>',
            help='Server ip domain.',
            required=True,
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Server port.',
            required=True,
            type=int,
        )
        parser.add_argument(
            '--weight',
            metavar='<weight>',
            help='server weight.',
            required=True,
            type=int,
        )
        parser.add_argument(
            '--fail-timeout',
            metavar='<fail-timeout>',
            help='Server fail timeout.',
            required=True,
            type=int,
        )
        return parser

    def get_data(self, parsed_args):
        url = '/edge/domains/{}/upstreams/{}/servers'.format(parsed_args.domain, parsed_args.upstream)
        payload = {
            "ip_domain": parsed_args.ip_domain,
            "port": parsed_args.port,
            "weight": parsed_args.weight,
            "fail_timeout": parsed_args.fail_timeout,
        }
        server = self.app.session.post(url, payload)
        return server


class ShowSSL(ShowOne):
    """Show ssl details."""
    schema = schemas.SSLDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='edge domain name ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/ssl'.format(parsed_args.domain)
        ssl = self.app.session.get(uri)
        return ssl


class UpdateSSL(ShowOne):
    """Update ssl."""
    schema = schemas.SSLDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='edge domain ID or name',
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            help='SSL type.',
            choices=("custom", "auto", ),
        )
        parser.add_argument(
            '--policy',
            metavar='<policy>',
            help='SSL policy.',
            choices=("normal", "strict",),
        )
        parser.add_argument(
            '--enable',
            metavar='<enable>',
            help='SSL enable.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--hsts',
            metavar='<hsts>',
            help='SSL hsts.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--https-redirect',
            metavar='<https-redirect>',
            help='SSL https redirect.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--ocsp-check',
            metavar='<ocsp-check>',
            help='SSL Ocsp',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--min-tls-versions',
            metavar='<min-tls-versions>',
            help='SSL Min TLS Versions.',
            choices=(
                'TLS_1_0',
                'TLS_1_1',
                'TLS_1_2',
                'TLS_1_3',
            ),
        )
        parser.add_argument(
            '--private-key',
            metavar='<private-key>',
            help='SSL private-key filename.',
        )
        parser.add_argument(
            '--public-key',
            metavar='<public-key>',
            help='SSL public-key filename.',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/ssl'.format(
            parsed_args.domain,
        )
        payload = {}
        if parsed_args.type:
            payload["type"] = parsed_args.type
        if parsed_args.policy:
            payload["policy"] = parsed_args.policy
        if parsed_args.enable is not None:
            payload["enable"] = parsed_args.enable
        if parsed_args.hsts is not None:
            payload["hsts"] = parsed_args.hsts
        if parsed_args.https_redirect is not None:
            payload["https_redirect"] = parsed_args.https_redirect
        if parsed_args.ocsp_check is not None:
            payload["ocsp_check"] = parsed_args.ocsp_check
        if parsed_args.min_tls_versions is not None:
            payload["min_tls_versions"] = parsed_args.min_tls_versions
        if parsed_args.public_key:
            with open(os.path.expanduser(parsed_args.public_key)) as file_:
                payload["public_key"] = file_.read()
        if parsed_args.private_key:
            with open(os.path.expanduser(parsed_args.private_key)) as file_:
                payload["private_key"] = file_.read()

        server = self.app.session.patch(uri, payload)
        return server


class ListLocation(Lister):
    """List edge locations"""
    schema = schemas.LocationDetails(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        locations = self.app.session.get('/edge/domains/{}/locations'.format(parsed_args.domain))
        return locations


class ShowLocation(ShowOne):
    """Show location details."""
    schema = schemas.LocationDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'location',
            metavar='<location>',
            help='edge location name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/locations/{}'.format(parsed_args.domain, parsed_args.location)
        location = self.app.session.get(uri)
        return location


class DeleteLocation(Command):
    """Delete location."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'location',
            metavar='<location>',
            help='edge location name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/edge/domains/{}/locations/{}'.format(parsed_args.domain, parsed_args.location)
        self.app.session.delete(uri)


class UpdateLocation(ShowOne):
    """Update location."""
    schema = schemas.LocationDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'location',
            metavar='<location>',
            help='edge location name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='upstream name.',
        )
        parser.add_argument(
            '--path',
            metavar='<path>',
            help='edge location path.',
        )
        parser.add_argument(
            '--path-case-sensitive',
            metavar='<path-case-sensitive>',
            help='path case sensitive enabled.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--upstream',
            metavar='<upstream>',
            help='location upstream.',
        )
        parser.add_argument(
            '--origin-headers',
            metavar='<origin-headers>',
            help='location origin headers.',
        )
        parser.add_argument(
            '--response-headers',
            metavar='<response-headers>',
            help='location response headers.',
        )
        parser.add_argument(
            '--cache-enabled',
            metavar='<cache-enabled>',
            help='location cache enabled.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--cache-ttl',
            metavar='<cache-ttl>',
            help='location cache time to live.',
            choices=("2m", "10m", "30m", "1h", "3h", "10h", "1d", "2d", "1w", "1M",),
        )
        parser.add_argument(
            '--cache-key',
            metavar='<cache-key>',
            help='location cache key.',
            choices=("u", "uq", "uqc",),
            type=str.lower,
        )
        parser.add_argument(
            '--cache-cookie-name',
            metavar='<cache-cookie-name>',
            help='location cache cookie name.',
        )
        parser.add_argument(
            '--cache-zone',
            metavar='<cache-zone>',
            help='location cache zone id.',
        )
        parser.add_argument(
            '--secure-link-enable',
            metavar='<secure-link-enable>',
            help='secure link enabled.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--secure-link-secret',
            metavar='<secure-link-secret>',
            help='secure link secret.',
        )
        parser.add_argument(
            '--redirect-enable',
            metavar='<redirect-enable>',
            help='redirect enabled.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--redirect-url',
            metavar='<redirect-url>',
            help='redirect url.',
        )
        parser.add_argument(
            '--redirect-status-code',
            metavar='<redirect-status-code>',
            help='redirect status code.',
            choices=("permanent", "temporary",),
            type=str.lower,
        )
        parser.add_argument(
            '--rate-limit',
            metavar='<rate-limit>',
            help='location rate limit id.',
        )
        return parser

    def get_data(self, parsed_args):
        redirect_status_code_dict = {
            "permanent": 301,
            "temporary": 302,
        }
        uri = '/edge/domains/{}/locations/{}'.format(parsed_args.domain, parsed_args.location)
        payload = {}

        if parsed_args.name:
            payload["name"] = parsed_args.name
        if parsed_args.path:
            payload["path"] = parsed_args.path
        if parsed_args.path_case_sensitive is not None:
            payload["path_case_sensitive"] = parsed_args.path_case_sensitive
        if parsed_args.upstream:
            payload["upstream_id"] = parsed_args.upstream
        if parsed_args.origin_headers:
            if str(parsed_args.origin_headers).lower() == "null":
                payload["origin_headers"] = []
            else:
                with open(os.path.expanduser(parsed_args.origin_headers)) as file_:
                    origin_headers_content = file_.read()
                    try:
                        payload["origin_headers"] = json.loads(origin_headers_content)
                    except json.decoder.JSONDecodeError as e:
                        raise CLIError(
                            [dict(msg="origin-headers is not a valid json")])
        if parsed_args.response_headers:
            if str(parsed_args.response_headers).lower() == "null":
                payload["response_headers"] = []
            else:
                with open(os.path.expanduser(parsed_args.response_headers)) as file_:
                    response_headers_content = file_.read()
                    try:
                        payload["response_headers"] = json.loads(response_headers_content)
                    except json.decoder.JSONDecodeError as e:
                        raise CLIError(
                            [dict(msg="response-headers is not a valid json")])
        if parsed_args.cache_enabled is not None:
            payload["cache_enabled"] = parsed_args.cache_enabled
        if parsed_args.cache_ttl:
            payload["cache_ttl"] = parsed_args.cache_ttl
        if parsed_args.cache_key:
            payload["cache_key"] = parsed_args.cache_key
        if parsed_args.cache_cookie_name:
            payload["cache_cookie_name"] = parsed_args.cache_cookie_name
        if parsed_args.cache_zone:
            payload["cache_zone"] = parsed_args.cache_zone
        if parsed_args.secure_link_enable is not None:
            payload["secure_link_enable"] = parsed_args.secure_link_enable
        if parsed_args.secure_link_secret:
            payload["secure_link_secret"] = parsed_args.secure_link_secret
        if parsed_args.redirect_enable is not None:
            payload["redirect_enable"] = parsed_args.redirect_enable
            if parsed_args.redirect_enable is True:
                if not parsed_args.redirect_url:
                    raise CLIError(
                        [dict(
                            msg="You need to specify redirect_url "
                                "when redirect_enable switch is True"
                        )]
                    )
                if not parsed_args.redirect_status_code:
                    payload["redirect_status_code"] = 302
        if parsed_args.redirect_url:
            payload["redirect_url"] = parsed_args.redirect_url
        if parsed_args.redirect_status_code:
            payload["redirect_status_code"] = redirect_status_code_dict[
                parsed_args.redirect_status_code
            ]
        if parsed_args.rate_limit:
            payload["rate_limit_id"] = parsed_args.rate_limit
        location = self.app.session.patch(uri, payload)
        return location


class CreateLocation(ShowOne):
    """Create location"""
    schema = schemas.LocationDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='upstream name.',
            required=True,
        )
        parser.add_argument(
            '--path',
            metavar='<path>',
            help='edge location path.',
            required=True,
        )
        parser.add_argument(
            '--path-case-sensitive',
            metavar='<path-case-sensitive>',
            help='path case sensitive enabled.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--upstream',
            metavar='<upstream>',
            help='location upstream id.',
            required=True,
        )
        parser.add_argument(
            '--origin-headers',
            metavar='<origin-headers>',
            help='location origin headers.',
        )
        parser.add_argument(
            '--response-headers',
            metavar='<response-headers>',
            help='location response headers.',
        )
        parser.add_argument(
            '--cache-enabled',
            metavar='<cache-enabled>',
            help='location cache enabled.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--cache-ttl',
            metavar='<cache-ttl>',
            help='location cache time to live.',
            choices=("2m", "10m", "30m", "1h", "3h", "10h", "1d", "2d", "1w", "1M",),
        )
        parser.add_argument(
            '--cache-key',
            metavar='<cache-key>',
            help='location cache key.',
            choices=("u", "uq", "uqc",),
            type=str.lower,
        )
        parser.add_argument(
            '--cache-cookie-name',
            metavar='<cache-cookie-name>',
            help='location cache cookie name.',
        )
        parser.add_argument(
            '--cache-zone',
            metavar='<cache-zone>',
            help='location cache zone id.',
        )
        parser.add_argument(
            '--secure-link-enable',
            metavar='<secure-link-enable>',
            help='secure link enabled.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--secure-link-secret',
            metavar='<secure-link-secret>',
            help='secure link secret.',
        )
        parser.add_argument(
            '--redirect-enable',
            metavar='<redirect-enable>',
            help='redirect enabled.',
            type=t_or_f_or_n,
        )
        parser.add_argument(
            '--redirect-url',
            metavar='<redirect-url>',
            help='redirect url.',
        )
        parser.add_argument(
            '--redirect-status-code',
            metavar='<redirect-status-code>',
            help='redirect status code.',
            choices=("permanent", "temporary",),
            type=str.lower,
        )
        parser.add_argument(
            '--rate-limit',
            metavar='<rate-limit>',
            help='location rate limit id.',
        )
        return parser

    def get_data(self, parsed_args):
        redirect_status_code_dict = {
            "permanent": 301,
            "temporary": 302,
        }
        url = '/edge/domains/{}/locations'.format(parsed_args.domain)
        payload = {
            "name": parsed_args.name,
            "path": parsed_args.path,
            "upstream_id": parsed_args.upstream,
            "redirect_enable": False,
            "redirect_status_code": 302,
        }
        if parsed_args.path_case_sensitive is not None:
            payload["path_case_sensitive"] = parsed_args.path_case_sensitive
        if parsed_args.origin_headers:
            with open(os.path.expanduser(parsed_args.origin_headers)) as file_:
                origin_headers_content = file_.read()
                try:
                    payload["origin_headers"] = json.loads(origin_headers_content)
                except json.decoder.JSONDecodeError as e:
                    raise CLIError(
                        [dict(msg="origin-headers is not a valid json")])
        if parsed_args.response_headers:
            with open(os.path.expanduser(parsed_args.response_headers)) as file_:
                response_headers_content = file_.read()
                try:
                    payload["response_headers"] = json.loads(response_headers_content)
                except json.decoder.JSONDecodeError as e:
                    raise CLIError(
                        [dict(msg="response-headers is not a valid json")])
        if parsed_args.cache_enabled is not None:
            payload["cache_enabled"] = parsed_args.cache_enabled
        if parsed_args.cache_ttl:
            payload["cache_ttl"] = parsed_args.cache_ttl
        if parsed_args.cache_key:
            payload["cache_key"] = parsed_args.cache_key
        if parsed_args.cache_cookie_name:
            payload["cache_cookie_name"] = parsed_args.cache_cookie_name
        if parsed_args.cache_zone:
            payload["cache_zone"] = parsed_args.cache_zone
        if parsed_args.secure_link_enable is not None:
            payload["secure_link_enable"] = parsed_args.secure_link_enable
        if parsed_args.secure_link_secret:
            payload["secure_link_secret"] = parsed_args.secure_link_secret
        if parsed_args.redirect_enable is not None:
            payload["redirect_enable"] = parsed_args.redirect_enable
            if parsed_args.redirect_enable is True:
                if parsed_args.redirect_url:
                    payload["redirect_url"] = parsed_args.redirect_url
                else:
                    raise CLIError(
                        [dict(
                            msg="You need to specify redirect_url "
                                "when redirect_enable switch is True"
                        )]
                    )
            if parsed_args.redirect_status_code:
                payload["redirect_status_code"] = redirect_status_code_dict[
                    parsed_args.redirect_status_code
                ]
        if parsed_args.rate_limit:
            payload["rate_limit_id"] = parsed_args.rate_limit
        location = self.app.session.post(url, payload)
        return location


class ListFirewall(Lister):
    """List edge firewalls"""
    schema = schemas.EdgeFirewallDetails(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        firewalls = self.app.session.get('/edge/domains/{}/firewalls'.format(parsed_args.domain))
        return firewalls


class ShowFirewall(ShowOne):
    """Show firewall details."""
    schema = schemas.EdgeFirewallDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'firewall',
            metavar='<firewall>',
            help='edge firewall ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/firewalls/{}'.format(parsed_args.domain, parsed_args.firewall)
        firewall = self.app.session.get(uri)
        return firewall


class DeleteFirewall(Command):
    """Delete firewall."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'firewall',
            metavar='<firewall>',
            help='edge firewall ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/edge/domains/{}/firewalls/{}'.format(parsed_args.domain, parsed_args.firewall)
        self.app.session.delete(uri)


class UpdateFirewall(ShowOne):
    """Update firewall."""
    schema = schemas.EdgeFirewallDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'firewall',
            metavar='<firewall>',
            help='edge firewall ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--input',
            metavar='<input>',
            help='firewall rule input type.',
            choices=("ip", "asn", "country", "continent",),
            type=str.lower,
        )
        parser.add_argument(
            '--value',
            metavar='<value>',
            help='firewall rule value path.',
        )
        parser.add_argument(
            '--action',
            metavar='<action>',
            help='firewall rule action.',
            choices=("allow", "block", "javascript_challenge", "captcha_challenge",),
            type=str.lower,
        )
        parser.add_argument(
            '--operator',
            metavar='<operator>',
            help='firewall rule operator',
            choices=("eq", "neq",),
            type=str.lower,
        )
        parser.add_argument(
            '--location',
            metavar='<location>',
            help='firewall location id.',
        )
        parser.add_argument(
            '--priority',
            metavar='<priority>',
            help='firewall rule priority.',
            type=int,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/firewalls/{}'.format(parsed_args.domain, parsed_args.firewall)
        payload = {}
        if parsed_args.input:
            payload["input"] = parsed_args.input
        if parsed_args.value:
            payload["value"] = parsed_args.value
        if parsed_args.action:
            payload["action"] = parsed_args.action
        if parsed_args.operator:
            payload["operator"] = parsed_args.operator
        if parsed_args.location:
            payload["location"] = parsed_args.location
        if parsed_args.priority:
            payload["priority"] = parsed_args.priority
        location = self.app.session.patch(uri, payload)
        return location


class CreateFirewall(ShowOne):
    """Create firewall"""
    schema = schemas.EdgeFirewallDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--input',
            metavar='<input>',
            help='firewall rule input type.',
            choices=("ip", "asn", "country", "continent",),
            type=str.lower,
            required=True,
        )
        parser.add_argument(
            '--value',
            metavar='<value>',
            help='firewall rule value path.',
            required=True,
        )
        parser.add_argument(
            '--action',
            metavar='<action>',
            help='firewall rule action.',
            choices=("allow", "block", "javascript_challenge", "captcha_challenge",),
            type=str.lower,
            required=True,
        )
        parser.add_argument(
            '--operator',
            metavar='<operator>',
            help='firewall rule operator',
            choices=("eq", "neq",),
            type=str.lower,
            required=True,
        )
        parser.add_argument(
            '--location',
            metavar='<location>',
            help='firewall location id.',
            required=True,
        )
        parser.add_argument(
            '--priority',
            metavar='<priority>',
            help='firewall rule priority.',
            type=int,
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        url = '/edge/domains/{}/firewalls'.format(parsed_args.domain)
        payload = {
            "input": parsed_args.input,
            "value": parsed_args.value,
            "action": parsed_args.action,
            "operator": parsed_args.operator,
            "location_id": parsed_args.location,
            "priority": parsed_args.priority,
        }
        location = self.app.session.post(url, payload)
        return location


class ListCacheZone(Lister):
    """List edge cache zones"""
    schema = schemas.CacheZoneList(many=True)

    def get_data(self, parsed_args):
        domains = self.app.session.get('/edge/cache-zones')
        return domains


class ShowGeneralSetting(ShowOne):
    """Show general setting details."""
    schema = schemas.GeneralSettingSchema()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='edge domain name ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/general-setting'.format(parsed_args.domain)
        general_setting = self.app.session.get(uri)
        return general_setting


class UpdateGeneralSetting(ShowOne):
    """Update general setting."""
    schema = schemas.GeneralSettingSchema()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='edge domain ID or name',
        )
        parser.add_argument(
            '--developer-mode',
            metavar='<developer-mode>',
            help='General setting developer mode.',
            type=t_or_f_or_n,
            required=False,
        )
        parser.add_argument(
            '--maintenance-mode',
            metavar='<maintenance-mode>',
            help='General setting maintenance mode.',
            type=t_or_f_or_n,
            required=False,
        )
        parser.add_argument(
            '--redirect-to-www',
            metavar='<redirect-to-www>',
            help='General setting redirect to www.',
            type=t_or_f_or_n,
            required=False,
        )
        parser.add_argument(
            '--ip-geolocation',
            metavar='<ip-geolocation>',
            help='General setting ip geo location.',
            type=t_or_f_or_n,
            required=False,
        )
        parser.add_argument(
            '--intercept-errors',
            metavar='<intercept_errors>',
            help='General setting intercepts errors.',
            type=t_or_f_or_n,
            required=False,
        )
        parser.add_argument(
            '--max-upload-size',
            metavar='<max-upload-size>',
            help='General setting max upload size.',
            type=int,
            required=False,
        )
        parser.add_argument(
            '--custom-host-header',
            metavar='<custom-host-header>',
            help='General setting custom host header.',
            type=str,
            required=False,
        )
        parser.add_argument(
            '--origin-headers',
            metavar='<origin-headers>',
            help='location origin headers.',
            required=False,
        )
        parser.add_argument(
            '--response-headers',
            metavar='<response-headers>',
            help='location response headers.',
            required=False,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/general-setting'.format(
            parsed_args.domain,
        )
        payload = {}
        if parsed_args.developer_mode is not None:
            payload["developer_mode"] = parsed_args.developer_mode
        if parsed_args.maintenance_mode is not None:
            payload["maintenance_mode"] = parsed_args.maintenance_mode
        if parsed_args.redirect_to_www is not None:
            payload["redirect_to_www"] = parsed_args.redirect_to_www
        if parsed_args.ip_geolocation is not None:
            payload["ip_geolocation"] = parsed_args.ip_geolocation
        if parsed_args.intercept_errors is not None:
            payload["intercept_errors"] = parsed_args.intercept_errors
        if parsed_args.max_upload_size is not None:
            payload["max_upload_size"] = parsed_args.max_upload_size
        if parsed_args.custom_host_header:
            payload["custom_host_header"] = parsed_args.custom_host_header
        if parsed_args.origin_headers:
            with open(os.path.expanduser(parsed_args.origin_headers)) as file_:
                origin_headers_content = file_.read()
                try:
                    payload["origin_headers"] = json.loads(origin_headers_content)
                except json.decoder.JSONDecodeError as e:
                    raise CLIError(
                        [dict(msg="origin-headers is not a valid json")])
        if parsed_args.response_headers:
            with open(os.path.expanduser(parsed_args.response_headers)) as file_:
                response_headers_content = file_.read()
                try:
                    payload["response_headers"] = json.loads(response_headers_content)
                except json.decoder.JSONDecodeError as e:
                    raise CLIError(
                        [dict(msg="response-headers is not a valid json")])
        server = self.app.session.patch(uri, payload)
        return server


class PurgeCache(Command):
    """purge cache"""
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help='edge domain ID or name',
        )
        parser.add_argument(
            '--url',
            metavar='<url>',
            help='url to purge cache',
            required=True,
        )
        parser.add_argument(
            '--method',
            metavar='<method>',
            help='Url method purge cache.',
            choices=("GET", "HEAD", "OPTIONAL",),
            type=str.upper,
            required=False,
        )
        parser.add_argument(
            '--cookies',
            metavar='<cookies>',
            help='purge cache url cookies path.',
            required=False,
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/edge/domains/{}/purge-cache'.format(
            parsed_args.domain,
        )
        payload = {}
        if parsed_args.url:
            payload["url"] = parsed_args.url
        if parsed_args.method:
            payload["method"] = parsed_args.method
        if parsed_args.cookies:
            with open(os.path.expanduser(parsed_args.cookies)) as file_:
                cookie_content = file_.read()
                try:
                    payload["cookies"] = json.loads(cookie_content)
                except json.decoder.JSONDecodeError as e:
                    raise CLIError(
                        [dict(msg="cookies is not a valid json")])
        server = self.app.session.post(uri, payload)
        return server


class ListRateLimit(Lister):
    """List edge rate limits"""
    schema = schemas.RateLimitDetails(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        rate_limits = self.app.session.get('/edge/domains/{}/rate-limits'.format(parsed_args.domain))
        return rate_limits


class ShowRateLimit(ShowOne):
    """Show rate limit details."""
    schema = schemas.RateLimitDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'rate_limit',
            metavar='<rate-limit>',
            help='edge rate limit name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain ID',
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/rate-limits/{}'.format(parsed_args.domain, parsed_args.rate_limit)
        rate_limit = self.app.session.get(uri)
        return rate_limit


class DeleteRateLimit(Command):
    """Delete rate limit."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'rate_limit',
            metavar='<rate-limit>',
            help='edge rate limit name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID',
            required=True,
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/edge/domains/{}/rate-limits/{}'.format(parsed_args.domain, parsed_args.rate_limit)
        self.app.session.delete(uri)


class UpdateRateLimit(ShowOne):
    """Update rate limit."""
    schema = schemas.RateLimitDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'rate_limit',
            metavar='<rate-limit>',
            help='edge rate limit name or ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='rate limit name.',
        )
        parser.add_argument(
            '--requests',
            metavar='<requests>',
            help='edge rate limit requests.',
            type=int,
        )
        parser.add_argument(
            '--burst',
            metavar='<burst>',
            help='rate limit burst type.',
            type=int,
        )
        parser.add_argument(
            '--exclusions',
            metavar='<exclusions>',
            help='List rate limit exclusions.',
            nargs='+',
        )
        parser.add_argument(
            '--time',
            metavar='<time>',
            help='rate limit time.',
            choices=("s", "m",),
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/edge/domains/{}/rate-limits/{}'.format(parsed_args.domain, parsed_args.rate_limit)
        payload = {}
        if parsed_args.name:
            payload["name"] = parsed_args.name
        if parsed_args.requests:
            payload["requests"] = parsed_args.requests
        if parsed_args.burst:
            payload["burst"] = parsed_args.burst
        if parsed_args.time:
            payload["time"] = parsed_args.time
        if parsed_args.exclusions:
            payload["exclusion_list"] = parsed_args.exclusions
        rate_limit = self.app.session.patch(uri, payload)
        return rate_limit


class CreateRateLimit(ShowOne):
    """Create rate limit"""
    schema = schemas.RateLimitDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='edge domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='rate limit name.',
            required=True,
        )
        parser.add_argument(
            '--requests',
            metavar='<requests>',
            help='edge rate limit requests.',
            type=int,
            required=True,
        )
        parser.add_argument(
            '--burst',
            metavar='<burst>',
            help='rate limit burst type.',
            type=int,
            required=True,
        )
        parser.add_argument(
            '--exclusions',
            metavar='<exclusions>',
            help='List rate limit exclusions.',
            nargs='+',
        )
        parser.add_argument(
            '--time',
            metavar='<time>',
            help='rate limit time.',
            choices=("s", "m",),
            required=True,
        )
        return parser

    def get_data(self, parsed_args):
        url = '/edge/domains/{}/rate-limits'.format(parsed_args.domain)
        payload = {
            "name": parsed_args.name,
            "requests": parsed_args.requests,
            "burst": parsed_args.burst,
            "time": parsed_args.time,
        }
        if parsed_args.exclusions:
            payload["exclusion_list"] = parsed_args.exclusions
        rate_limit = self.app.session.post(url, payload)
        return rate_limit



from digicloud.commands.base import Lister, ShowOne
from digicloud.error_handlers import CLIError


class ListRegions(Lister):
    """List available regions"""

    def get_data(self, parsed_args):
        return self.app.session.get('/regions')


class SelectRegion(ShowOne):
    """Select specific region to setup"""

    def get_parser(self, prog_name):
        parser = super(SelectRegion, self).get_parser(prog_name)

        parser.add_argument(
            'region_name',
            metavar='<region_name>',
            help="Region name"
        )

        return parser

    def get_data(self, parsed_args):
        regions = {region['name'] for region in self.app.session.get('/regions')}
        if parsed_args.region_name not in regions:
            raise CLIError([dict(
                msg="We have no such a region in Digicloud",
                hint='Check our current regions via '
                     '[blue bold]digicloud region list[/blue bold]'
            )])
        self.app.config['AUTH_HEADERS'].update({
            'Digicloud-Region': parsed_args.region_name,
        })
        return {'Region': parsed_args.region_name}


class CurrentRegion(ShowOne):
    """Show current selected region"""

    def get_data(self, parsed_args):
        header = self.app.config.get('AUTH_HEADERS')
        if header is not None:
            selected_region = header.get('Digicloud-Region', 'No region selected')
        else:
            selected_region = 'No region selected'

        return {'Region': selected_region}



from digicloud.commands.base import Command
from digicloud.error_handlers import CLIError


class InstanceOperationMixin:
    new_state = None

    def get_parser(self, prog_name):
        parser = super(InstanceOperationMixin, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        return parser

    def _on_409(self, parsed_args, response):
        instance = self.app.session.get('/instances/%s' % parsed_args.instance)
        return CLIError([
            dict(
                msg="Your instance status is {}, and your requested operation "
                    "is not possible at the moment.".format(instance['status']),
                hint="Check your instance current status by "
                     "[blue bold]digicloud instance show {}[/blue bold]".format(
                    instance['name']
                )
            )
        ]
        )

    def take_action(self, parsed_args):
        uri = '/instances/%s/state-transitions' % parsed_args.instance
        self.app.session.post(uri, payload=dict(new_state=self.new_state))


class StartInstance(InstanceOperationMixin, Command):
    """Start instance."""
    new_state = "ACTIVE"


class StopInstance(InstanceOperationMixin, Command):
    """Stop instance."""
    new_state = "SHUTOFF"


class SuspendInstance(InstanceOperationMixin, Command):
    """Suspend instance."""
    new_state = "SUSPENDED"


class ResumeInstance(InstanceOperationMixin, Command):
    """Resume instance."""
    new_state = "ACTIVE"


class RebootInstance(InstanceOperationMixin, Command):
    """Reboot instance."""

    def get_parser(self, prog_name):
        parser = super(RebootInstance, self).get_parser(prog_name)
        parser.add_argument(
            '--type',
            default='SOFT',
            metavar='<type>',
            help='either `SOFT` for a software-level reboot, or `HARD` for a virtual '
                 'power cycle hard reboot',
            choices=['SOFT', 'HARD']
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/state-transitions' % parsed_args.instance
        self.app.session.post(uri, payload=dict(new_state="REBOOT",
                                                params={"reboot_type": parsed_args.type}))



"""
    DigiCloud InstanceTypes Service.
"""
from digicloud import schemas
from .base import Lister, ShowOne


class ListInstanceType(Lister):
    """List instance-types."""
    schema = schemas.InstanceType(many=True)

    def get_parser(self, prog_name):
        parser = super(ListInstanceType, self).get_parser(prog_name)
        parser.add_argument(
            '--family',
            metavar='<family>',
            help='Filter by family',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/instance-types'
        if parsed_args.family:
            uri = '/instance-types?family=%s' % parsed_args.family
        data = self._sort_instance_types(self.app.session.get(uri))
        return data

    def _sort_instance_types(self, instance_types):
        return sorted(
            instance_types,
            key=lambda item: (
                # Sorting keys
                item.get('family'),
                item.get('vcpus'),
                item.get('ram'),
            ),
            reverse=True
        )


class ShowInstanceType(ShowOne):
    """Show instance-type details."""
    schema = schemas.InstanceType()

    def get_parser(self, prog_name):
        parser = super(ShowInstanceType, self).get_parser(prog_name)
        parser.add_argument(
            'instance_type',
            metavar='<instance type>',
            help='InstanceType name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = f'/instance-types/{parsed_args.instance_type}'
        data = self.app.session.get(uri)
        return data



from .base import Lister, ShowOne
from .. import schemas
from ..error_handlers import CLIError


class QuotaRequestList(Lister):
    """List your namespace quota"""
    schema = schemas.QuotaRequest(many=True)

    def get_data(self, parsed_args):
        return self.app.session.get('/quota-requests')[:15]


class RequestMoreQuota(ShowOne):
    """Submit a request for more quota"""
    schema = schemas.QuotaRequest()

    def get_parser(self, prog_name):
        parser = super(RequestMoreQuota, self).get_parser(prog_name)
        parser.add_argument(
            '--quota-id',
            required=True,
            metavar='<quota_id>',
            help='Quota ID for the request',
        )

        parser.add_argument(
            '--value',
            required=True,
            metavar='<value>',
            help='Your desire value'
        )
        parser.add_argument(
            '--rule-index',
            required=False,
            type=int,
            metavar='<rule_index>',
            help='Your desire rule id'
        )

        return parser

    def get_data(self, parsed_args):
        resolution = self._check_rule_resolution(parsed_args.quota_id, parsed_args.rule_index)
        return self.app.session.post(
            '/quota-requests',
            {
                'quota_id': parsed_args.quota_id,
                'required_quota': parsed_args.value,
                'note': "Requested by CLI",
                'resolution': resolution,
            })

    def _check_rule_resolution(self, quota_id, rule_id):
        keys_map = {
            'edge.namespace.domain.record.count': 'domain_id',
        }
        uri = '/quotas/{}'.format(self.app.config['CURRENT']['Namespace ID'])
        quota_list = self.app.session.get(uri)
        related_quotas = [quota for quota in quota_list if quota['id'] == quota_id]
        related_quotas_len = len(related_quotas)
        if related_quotas_len == 1:
            return related_quotas[0]['resolution']
        elif related_quotas_len == 0:
            raise CLIError([
                dict(
                    msg="The quota id {} does not exist".format(quota_id)
                )
            ]
            )

        else:
            rule_key = keys_map.get(quota_id, None)
            if rule_key is None:
                raise CLIError([
                    dict(
                        msg="The quota id {} does not support multi rule".format(quota_id)
                    )
                ]
                )
            related_quotas.sort(key=lambda x: x['resolution'][rule_key])
            if rule_id and rule_id <= related_quotas_len:
                return related_quotas[rule_id-1]['resolution']
            else:
                errors = [
                    dict(
                        msg="Rule index {}: {}".format(index, rule['description'])
                    ) for index, rule in enumerate(related_quotas, 1)
                ]
                errors[-1]["hint"] = "You should choose one of the rules index provided above"
                raise CLIError(errors)




from os import path

from rich.markdown import Markdown

import digicloud
from digicloud.commands.base import Command
from digicloud.documentations import DocumentationContainer


class Explain(Command):
    """Digicloud manual for different topics"""

    def get_parser(self, prog_name):
        parser = super(Explain, self).get_parser(prog_name)
        parser.add_argument(
            'topic',
            metavar='<topic>',
            nargs="?",
            default=None,
            help='Topic to explain'
        )
        return parser

    def take_action(self, parsed_args):
        documentation = DocumentationContainer(
            path.join(path.dirname(digicloud.__file__), 'documentations')
        )
        if parsed_args.topic is None:
            self._display_list_of_topics(documentation.list())
            return
        try:

            self.app.console.print(
                Markdown(documentation.get(parsed_args.topic))
            )
        except FileNotFoundError:
            self.app.console.print(
                (
                    "[red bold]There is not documentation "
                    "available for '{}'.[/red bold]"
                ).format(
                    parsed_args.topic
                )
            )
            self._display_list_of_topics(documentation.list())

    def _display_list_of_topics(self, topics):
        self.app.console.print(
            "[green bold]The following "
            "topics are available:[/green bold]"
        )
        for topic in topics:
            self.app.console.print("* {}".format(topic))



import argparse

"""
Copied from osc-lib. If we need more tools from this library, we should install it.
"""


class MultiKeyValueAction(argparse.Action):
    """A custom action to parse arguments as key1=value1,key2=value2 pairs

    Ensure that ``dest`` is a list. The list will finally contain multiple
    dicts, with key=value pairs in them.

    NOTE: The arguments string should be a comma separated key-value pairs.
    And comma(',') and equal('=') may not be used in the key or value.
    """

    def __init__(self, option_strings, dest, nargs=None,
                 required_keys=None, optional_keys=None, **kwargs):
        """Initialize the action object, and parse customized options

        Required keys and optional keys can be specified when initializing
        the action to enable the key validation. If none of them specified,
        the key validation will be skipped.

        :param required_keys: a list of required keys
        :param optional_keys: a list of optional keys
        """
        if nargs:
            msg = "Parameter 'nargs' is not allowed, but got %s"
            raise ValueError(msg % nargs)

        super(MultiKeyValueAction, self).__init__(option_strings,
                                                  dest, **kwargs)

        # required_keys: A list of keys that is required. None by default.
        if required_keys and not isinstance(required_keys, list):
            msg = "'required_keys' must be a list"
            raise TypeError(msg)
        self.required_keys = set(required_keys or [])

        # optional_keys: A list of keys that is optional. None by default.
        if optional_keys and not isinstance(optional_keys, list):
            msg = "'optional_keys' must be a list"
            raise TypeError(msg)
        self.optional_keys = set(optional_keys or [])

    def __call__(self, parser, namespace, values, metavar=None):
        # Make sure we have an empty list rather than None
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, [])

        params = {}
        for kv in values.split(','):
            # Add value if an assignment else raise ArgumentTypeError
            if '=' in kv:
                kv_list = kv.split('=', 1)
                # NOTE(qtang): Prevent null key setting in property
                if '' == kv_list[0]:
                    msg = "Each property key must be specified: %s"
                    raise argparse.ArgumentTypeError(msg % str(kv))
                else:
                    params.update([kv_list])
            else:
                msg = "Expected comma separated 'key=value' pairs, but got: %s"
                raise argparse.ArgumentTypeError(msg % str(kv))

        # Check key validation
        valid_keys = self.required_keys | self.optional_keys
        if valid_keys:
            invalid_keys = [k for k in params if k not in valid_keys]
            if invalid_keys:
                msg = "Invalid keys %(invalid_keys)s specified.\n " \
                      "Valid keys are: %(valid_keys)s"

                raise argparse.ArgumentTypeError(msg % {
                    'invalid_keys': ', '.join(invalid_keys),
                    'valid_keys': ', '.join(valid_keys),
                })

        if self.required_keys:
            missing_keys = [k for k in self.required_keys if k not in params]
            if missing_keys:
                msg = "Missing required keys %(missing_keys)s.\n" \
                      "Required keys are: %(required_keys)s"

                raise argparse.ArgumentTypeError(msg % {
                    'missing_keys': ', '.join(missing_keys),
                    'required_keys': ', '.join(self.required_keys),
                })

        # Update the dest dict
        getattr(namespace, self.dest, []).append(params)


"""
    DigiCloud Network Service.
"""

from .base import Lister, ShowOne, Command
from .. import schemas
from ..error_handlers import CLIError


def enrich_network_details(session, *networks):
    subnets = {
        subnet['id']: subnet
        for subnet in session.get('/subnets')
    }
    for network_details in networks:
        network_details['subnets'] = [
            {
                "name": subnets[subnet_id]['name'],
                "cidr": subnets[subnet_id]['cidr'],
                "gateway": subnets[subnet_id]['gateway_ip'],
            }
            for subnet_id in network_details['subnets']
        ]


class ListNetwork(Lister):
    """List networks"""
    schema = schemas.NetworkList(many=True)

    def get_data(self, parsed_args):
        networks = self.app.session.get('/networks')
        enrich_network_details(self.app.session, *networks)
        return networks


class ShowNetwork(ShowOne):
    """Show network details."""
    schema = schemas.NetworkDetail()

    def get_parser(self, prog_name):
        parser = super(ShowNetwork, self).get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar='<network>',
            help='Network name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/networks/%s' % parsed_args.network
        network = self.app.session.get(uri)
        enrich_network_details(self.app.session, network)
        return network


class DeleteNetwork(Command):
    """Delete network."""

    def get_parser(self, prog_name):
        parser = super(DeleteNetwork, self).get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar='<network>',
            help='Network name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/networks/%s' % parsed_args.network
        self.app.session.delete(uri)


class UpdateNetwork(ShowOne):
    """Update network."""
    schema = schemas.NetworkDetail()

    def get_parser(self, prog_name):
        parser = super(UpdateNetwork, self).get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar='<network>',
            help='Network ID',
        )
        parser.add_argument(
            '--name',
            metavar='<Name>',
            help='New name for network.'
        )
        parser.add_argument(
            '--description',
            metavar='<Description>',
            help='Description fro network.'
        )
        parser.add_argument(
            '--admin-state',
            metavar='<admin_state>',
            help='New admin state.',
            choices=("UP", "DOWN"),
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/networks/%s' % parsed_args.network
        payload = {}
        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.description:
            payload['description'] = parsed_args.description
        if parsed_args.admin_state:
            payload['admin_state'] = parsed_args.admin_state
        if not payload:
            raise CLIError([dict(
                msg="At least one of --name or --admin-state is necessary"
            )])
        network = self.app.session.patch(uri, payload)
        enrich_network_details(self.app.session, network)
        return network


class CreateNetwork(ShowOne):
    """Create Network"""
    schema = schemas.NetworkDetail()

    def get_parser(self, prog_name):
        parser = super(CreateNetwork, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Network name'
        )
        parser.add_argument(
            '--description',
            metavar='<Description>',
            help='Network description'
        )
        parser.add_argument(
            '--admin-state',
            metavar='<AdminState>',
            choices=("UP", "DOWN"),
            default='UP',
            help='Set admin state, could be "UP" or "DOWN", (default UP).'
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            'name': parsed_args.name,
            'admin_state': parsed_args.admin_state
        }
        if parsed_args.description:
            payload['description'] = parsed_args.description
        network = self.app.session.post('/networks', payload)
        return network



"""
    Authentication to digicloud API.
"""

from rich.prompt import Prompt
from .. import schemas
from digicloud.commands.base import ShowOne, Command, Lister
from digicloud.error_handlers import CLIError
from digicloud.utils import is_tty


class Login(ShowOne):
    """Digicloud Login"""

    def get_parser(self, prog_name):
        parser = super(Login, self).get_parser(prog_name)
        parser.add_argument(
            '--email',
            required=False,
            default=None,
            metavar='<EMAIL>',
            dest='email',
            help=("Your Digicloud email")
        )

        parser.add_argument(
            '--password',
            metavar='<PASSWORD>',
            required=False,
            default=None,
            dest='password',
            help=("Your Digicloud account password")
        )

        parser.add_argument(
            '--namespace',
            metavar='<NAMESPACE>',
            dest='namespace',
            help=("ID of the Digicloud namespace")
        )

        return parser

    def get_data(self, parsed_args):
        payload = {
            'email': parsed_args.email or self._get_email(),
            'password': parsed_args.password or self._get_password(),
        }
        data = self.app.session.post('/tokens', payload)

        # NOTE: User may have no default namespace
        ns_ref = parsed_args.namespace or data['user']['default_namespace']
        selected_ns = self._find_namespace(ns_ref, data['namespaces'])

        self.app.config['api_token'] = data['token']
        self.app.config['AUTH_HEADERS'].update({
            'Authorization': 'Bearer {}'.format(data["token"]),
            'Digicloud-Namespace': selected_ns['id'],
        })
        self.app.config['USER'] = data['user']
        self.app.config['CURRENT'] = current_account = {
            "Email": data['user']['email'],
            "Namespace name": selected_ns['name'],
            "Namespace ID": selected_ns['id'],
            "Region": self._get_active_region()
        }
        return current_account

    def _find_namespace(self, name_or_id, namespaces):
        if name_or_id is None:
            return namespaces[0]

        result = [ns for ns in namespaces
                  if name_or_id == ns['name'] or name_or_id == ns['id']]
        if len(result) == 0:
            print("You're not a member of '{}' namespace".format(name_or_id))
            return
        elif len(result) > 1:
            print("There are more than one namespace named '{}'".format(name_or_id))
            return
        return result[0]

    def _get_active_region(self):
        self.app.session.setup(self.app.config)
        auth_headers = self.app.config['AUTH_HEADERS']
        if 'Digicloud-Region' not in auth_headers:
            regions = self.app.session.get('/regions')
            if len(regions) == 0:
                raise CLIError([
                    dict(msg="There are no region available for your namespace, "
                             "Please contact support for further information.")
                ])
            auth_headers['Digicloud-Region'] = regions[0]['name']
        return auth_headers['Digicloud-Region']

    def _get_email(self):
        if not is_tty():
            raise CLIError([
                dict(msg="Unable to login in non-interactive mode without --email")
            ])
        return Prompt.ask("Email")

    def _get_password(self):
        if not is_tty():
            raise CLIError([
                dict(
                    msg="Unable to login in non-interactive mode without --password"
                )
            ])
        return Prompt.ask("Password", password=True)

    def _on_401(self, parsed_args, exp):
        return CLIError([
            dict(
                msg="Username and password do not matched",
                hint="use this link in case you forget your password: "
                     "https://console.digicloud.ir/auth/forgot-password"
            )
        ])


class Logout(Command):
    """DigiCloud Logout"""

    def take_action(self, parsed_args):
        if 'api_token' in self.app.config:
            self.app.session.delete('/tokens/{}'.format(self.app.config['api_token']))
            del self.app.config['api_token']

        del self.app.config['AUTH_HEADERS']
        del self.app.session.headers['Authorization']
        del self.app.config['CURRENT']

        return 0


class CurrentUser(ShowOne):
    """Displays current logged in user"""

    def get_data(self, parsed_args):
        if 'CURRENT' in self.app.config:
            return self.app.config['CURRENT']
        raise CLIError([{
            "msg": "You're not logged in"
        }])



from cliff.formatters import base
from rich import box
from rich.console import Console
from rich.table import Table, Column


class RichFormatter(base.ListFormatter, base.SingleFormatter):
    def add_argument_group(self, parser):
        pass

    def emit_list(self, column_names, data, stdout, parsed_args):
        table = Table(
            *[Column(c, overflow='fold') for c in column_names],
            width=int(parsed_args.max_width) or None,
        )
        for row in data:
            table.add_row(*[str(cell) for cell in row])
        Console(emoji=False, markup=False).print(table)

    def emit_one(self, column_names, data, stdout, parsed_args):
        table = Table(
            Column("Key", overflow='fold'),
            Column("Value", overflow='fold'),
            show_header=False,
            box=box.SQUARE,
        )
        for key, value in list(zip(column_names, data)):
            table.add_row(key, str(value))
        Console(emoji=False, markup=False).print(table)



"""
    DigiCloud Access Service.
"""

from .base import Lister, ShowOne, Command
from .. import schemas


class ListPublicIp(Lister):
    """List public Ips."""
    schema = schemas.PublicIPList(many=True)

    def get_data(self, parsed_args):
        uri = '/public-ips'
        return self.app.session.get(uri)


class ShowPublicIp(ShowOne):
    """Show public Ip details."""
    schema = schemas.PublicIPDetails()

    def get_parser(self, prog_name):
        parser = super(ShowPublicIp, self).get_parser(prog_name)
        parser.add_argument(
            'public_ip',
            metavar='<public_ip>',
            help=('Public IP ID'),
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/public-ips/%s' % parsed_args.public_ip
        return self.app.session.get(uri)


class CreatePublicIp(ShowOne):
    """Create Public Ip."""
    schema = schemas.PublicIPDetails()

    def get_data(self, parsed_args):
        return self.app.session.post('/public-ips', {})


class DeletePublicIp(Command):
    """Delete Public Ip."""

    def get_parser(self, prog_name):
        parser = super(DeletePublicIp, self).get_parser(prog_name)
        parser.add_argument(
            'public_ip',
            metavar='<public_ip>',
            help=('Public IP ID')
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/public-ips/%s' % parsed_args.public_ip
        self.app.session.delete(uri)


class AssociatePublicIp(ShowOne):
    """Associate Public Ip."""
    schema = schemas.PublicIPDetails()

    def get_parser(self, prog_name):
        parser = super(AssociatePublicIp, self).get_parser(prog_name)
        parser.add_argument(
            'public_ip_id',
            metavar='<public_ip_id>',
            help='Public IP ID'
        )
        parser.add_argument(
            '--interface-id',
            metavar='<interface_id>',
            required=True,
            help='Interface ID'
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            'interface_id': parsed_args.interface_id,
        }
        uri = '/public-ips/%s/associate' % parsed_args.public_ip_id
        return self.app.session.post(uri, payload)


class RevokePublicIp(ShowOne):
    """Revoke Public Ip."""
    schema = schemas.PublicIPDetails()

    def get_parser(self, prog_name):
        parser = super(RevokePublicIp, self).get_parser(prog_name)
        parser.add_argument(
            'public_ip_id',
            metavar='<public_ip_id>',
            help='Public IP ID'
        )

        return parser

    def get_data(self, parsed_args):
        uri = '/public-ips/%s/revoke' % parsed_args.public_ip_id
        return self.app.session.post(uri, {})



"""
    DigiCloud Router Service.
"""

from digicloud import schemas
from .base import Lister, ShowOne, Command
from ..error_handlers import CLIError


def enrich_router_interface_details(session, *routers):
    subnets = {
        subnet['id']: subnet['name']
        for subnet in session.get('/subnets')
    }
    networks = {
        network['id']: network['name']
        for network in session.get('/networks')
    }

    for router in routers:
        router['network_name'] = networks.get(router['network_id'], "N/A")
        router['subnet_name'] = subnets.get(router['subnet_id'], "N/A")


class ListRouter(Lister):
    """List Routers."""
    schema = schemas.RouterList(many=True)

    def get_data(self, parsed_args):
        data = self.app.session.get('/routers')

        return data


class ShowRouter(ShowOne):
    """Show Router details."""
    schema = schemas.RouterDetails(many=False)

    def get_parser(self, prog_name):
        parser = super(ShowRouter, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/routers/%s' % parsed_args.router
        data = self.app.session.get(uri)

        return data


class DeleteRouter(Command):
    """Delete router."""

    def get_parser(self, prog_name):
        parser = super(DeleteRouter, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s' % parsed_args.router
        self.app.session.delete(uri)


class CreateRouter(ShowOne):
    """Create Router."""
    schema = schemas.RouterDetails(many=False)

    def get_parser(self, prog_name):
        parser = super(CreateRouter, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Router name'
        )
        parser.add_argument(
            '--admin-state',
            metavar='<AdminState>',
            choices=['UP', 'DOWN'],
            type=str.upper,
            help='Set admin state, could be "UP" or "DOWN", default "UP"',
            default='UP',
        )

        return parser

    def get_data(self, parsed_args):
        payload = {
            'name': parsed_args.name,
            'admin_state': parsed_args.admin_state,
        }
        data = self.app.session.post('/routers', payload)
        return data


class ListRouterInterface(Lister):
    """List all router interfaces"""
    schema = schemas.RouterInterfaceList(many=True)

    def get_parser(self, prog_name):
        parser = super(ListRouterInterface, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = f'/routers/{parsed_args.router}/interfaces'
        interfaces = self.app.session.get(uri)
        enrich_router_interface_details(self.app.session, *interfaces)
        return interfaces


class ShowRouterInterface(ShowOne):
    """Show router interface."""
    schema = schemas.RouterInterfaceDetails(many=False)

    def get_parser(self, prog_name):
        parser = super(ShowRouterInterface, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        parser.add_argument(
            '--interface-id',
            required=True,
            metavar='<interface_id>',
            help='Interface ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/routers/%s/interfaces/%s' % (parsed_args.router, parsed_args.interface_id)
        interface = self.app.session.get(uri)
        enrich_router_interface_details(self.app.session, interface)
        return interface


class AddRouterInterface(ShowOne):
    """Add router interface."""
    schema = schemas.RouterInterfaceDetails(many=False)

    def get_parser(self, prog_name):
        parser = super(AddRouterInterface, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        parser.add_argument(
            '--subnet',
            required=True,
            metavar='<subnet>',
            help='SubnetList name or ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/routers/%s/interfaces' % parsed_args.router
        payload = {'subnet': parsed_args.subnet}
        interface = self.app.session.post(uri, payload)
        enrich_router_interface_details(self.app.session, interface)
        return interface


class RemoveRouterInterface(Command):
    """Remove router interface."""
    def get_parser(self, prog_name):
        parser = super(RemoveRouterInterface, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        parser.add_argument(
            '--interface-id',
            required=True,
            metavar='<interface_id>',
            help='Interface ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s/interfaces/%s' % (parsed_args.router, parsed_args.interface_id)
        self.app.session.delete(uri)


class AddRouterExternal(ShowOne):
    """Add external network to router."""
    schema = schemas.RouterDetails(many=False)

    def get_parser(self, prog_name):
        parser = super(AddRouterExternal, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/routers/%s' % parsed_args.router
        payload = {'has_gateway': True}
        data = self.app.session.patch(uri, payload)
        return data


class RemoveRouterExternal(ShowOne):
    """Remove external network from router."""
    schema = schemas.RouterDetails(many=False)

    def get_parser(self, prog_name):
        parser = super(RemoveRouterExternal, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/routers/%s' % parsed_args.router
        payload = {'has_gateway': False}
        data = self.app.session.patch(uri, payload)
        return data


class UpdateRouter(ShowOne):
    _description = "Update Router"
    schema = schemas.RouterDetails(many=False)

    def get_parser(self, prog_name):
        parser = super(UpdateRouter, self).get_parser(prog_name)
        group = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='New name for the router'
        )
        parser.add_argument(
            '--admin-state',
            metavar='<AdminState>',
            choices=['UP', 'DOWN'],
            type=str.upper,
            help='Set admin state, could be "UP" or "DOWN"'
        )

        group.add_argument(
            '--disable-gateway',
            help='Disconnect router from external network',
            action='store_true'

        )
        group.add_argument(
            '--enable-gateway',
            help='Connect router to external network',
            action='store_true'
        )
        return parser

    def get_data(self, parsed_args):
        payload = {}
        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.disable_gateway:
            payload['has_gateway'] = False
        if parsed_args.enable_gateway:
            payload['has_gateway'] = True
        if parsed_args.admin_state:
            payload['admin_state'] = parsed_args.admin_state
        if len(payload) == 0:
            raise CLIError([dict(
                msg="You need to at least provide one of "
                    "--name, --admin-state, --enable-gateway or --disable-gateway"
            )])
        uri = '/routers/%s' % parsed_args.router
        data = self.app.session.patch(uri, payload)
        return data


class ListRouterStatic(Lister):
    """List router static routes."""

    def get_parser(self, prog_name):
        parser = super(ListRouterStatic, self).get_parser(prog_name)
        parser.add_argument(
            'router_id',
            metavar='<router_id>',
            help='Router ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/routers/%s/static-routes' % parsed_args.router_id
        data = self.app.session.get(uri)
        return data


class AddRouterStatic(Lister):
    """Add router static routes."""
    def get_parser(self, prog_name):
        parser = super(AddRouterStatic, self).get_parser(prog_name)
        parser.add_argument(
            'router_id',
            metavar='<router_id>',
            help='Router ID'
        )
        parser.add_argument(
            '--destination',
            metavar='<Network>',
            required=True,
            help='Destination network.'
        )
        parser.add_argument(
            '--nexthop',
            metavar='<IP address>',
            required=True,
            help='Next hop IP address.'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/routers/%s/static-routes' % parsed_args.router_id
        payload = {
            "destination": parsed_args.destination,
            "nexthop": parsed_args.nexthop,

        }
        data = self.app.session.post(uri, payload)
        return data


class DeleteRouterStatic(Lister):
    """Delete static routes."""
    def get_parser(self, prog_name):
        parser = super(DeleteRouterStatic, self).get_parser(prog_name)
        parser.add_argument(
            'router_id',
            metavar='<router_id>',
            help='Router ID'
        )
        parser.add_argument(
            '--destination',
            metavar='<Network>',
            required=True,
            help='Destination network.'
        )
        parser.add_argument(
            '--nexthop',
            metavar='<IP address>',
            required=True,
            help='Next hop IP address.'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/routers/%s/static-routes' % parsed_args.router_id
        payload = {
            "destination": parsed_args.destination,
            "nexthop": parsed_args.nexthop,

        }
        data = self.app.session.delete(uri, payload)
        return data



class HumanBytes:
    METRIC_LABELS = ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    BINARY_LABELS = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]
    PRECISION_OFFSETS = [0.5, 0.05, 0.005, 0.0005]  # PREDEFINED FOR SPEED.
    PRECISION_FORMATS = ["{}{:.0f} {}", "{}{:.1f} {}", "{}{:.2f} {}",
                         "{}{:.3f} {}"]  # PREDEFINED FOR SPEED.

    @staticmethod
    def format(num, metric=False, precision=1, no_unit=False):
        """
        Human-readable formatting of bytes, using binary (powers of 1024)
        or metric (powers of 1000) representation.
        """

        assert isinstance(num, (int, float)), "num must be an int or float"
        assert isinstance(metric, bool), "metric must be a bool"
        assert isinstance(precision, int) and \
               0 <= precision <= 3, "precision must be an int (range 0-3)"

        unit_labels = HumanBytes.METRIC_LABELS if metric else HumanBytes.BINARY_LABELS
        last_label = unit_labels[-1]
        unit_step = 1000 if metric else 1024
        unit_step_thresh = unit_step - HumanBytes.PRECISION_OFFSETS[precision]

        is_negative = num < 0
        if is_negative:  # Faster than ternary assignment or always running abs().
            num = abs(num)

        for unit in unit_labels:
            if num < unit_step_thresh:
                # VERY IMPORTANT:
                # Only accepts the CURRENT unit if we're BELOW the threshold where
                # float rounding behavior would place us into the NEXT unit: F.ex.
                # when rounding a float to 1 decimal, any number ">= 1023.95" will
                # be rounded to "1024.0". Obviously we don't want ugly output such
                # as "1024.0 KiB", since the proper term for that is "1.0 MiB".
                break
            if unit != last_label:
                # We only shrink the number if we HAVEN'T reached the last unit.
                # NOTE: These looped divisions accumulate floating point rounding
                # errors, but each new division pushes the rounding errors further
                # and further down in the decimals, so it doesn't matter at all.
                num /= unit_step

        return HumanBytes.PRECISION_FORMATS[precision] \
            .format("-" if is_negative else "", num, "" if no_unit else unit)


def format_ram(value):
    return value / 1024


def format_size(value, multiplier=None, no_unit=False):
    multiplier = multiplier or (1000 ** 3)
    return HumanBytes.format(value * multiplier, True, no_unit=no_unit)

# TODO maybe need to create general class for formatting



from requests import HTTPError
from sentry_sdk import capture_exception

from .schemas import ErrorSchema
from .utils import convert_camel_to_snake


class CLIError(Exception):
    def __init__(self, errors, return_code=1):
        self.errors = errors
        self.return_code = return_code


class ErrorHandler:
    def __init__(self, app):
        self.app = app

    def handle(self, exception):
        if getattr(self.app.options, 'debug', False):
            raise exception
        try:
            if isinstance(exception, HTTPError):
                result = self._handle_error_code(exception.response)
                if result:
                    return result
            capture_exception(exception)
            return self.get_unexpected_error()
        except Exception as exp:
            capture_exception(exp)
            return self.get_unexpected_error()

    def _handle_error_code(self, response):
        error = ErrorSchema().load(response.json())
        handler_method = getattr(self, "_on_{}".format(
            convert_camel_to_snake(error["error_code"])
        ), self._on_not_supported_error_code)
        return handler_method(error)

    def _on_not_found(self, error):
        return CLIError([dict(
            msg=error["message"],
            hint="It might be a typo in your object name or ID",
        )])

    def _on_no_unique_match(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def _on_bad_request(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def _on_invalid_json(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def _on_conflict(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def _on_quota_exceeded(self, error):
        cli_errors = []
        for rule in error["rules"]:
            if rule["used"].lower() != "none":
                message = "{message} {description} is {quota} but you used {used}.".format(
                        message=error['message'],
                        description=rule['description'],
                        quota=rule['quota'],
                        used=rule['used'],
                    )
            else:
                message = "{message} {description} is {quota}.".format(
                        message=error['message'],
                        description=rule['description'],
                        quota=rule['quota'],
                    )
            cli_errors.append(
                dict(
                    msg=message,
                    hint="Request more quota.",
                )
            )
        return CLIError(cli_errors)

    def _on_unauthorized(self, error):
        return CLIError([dict(
            msg=error["message"],
            hint="You can login to you account via `digicloud account login`"
                 " or go to https://digicloud.ir for registration",
        )])

    def _on_forbidden(self, error):
        return CLIError([dict(
            msg=error["message"],
            hint="You might not have required permission",
        )])

    def _on_not_supported_error_code(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def _on_validation_error(self, error):
        return CLIError([dict(
            msg="{message} {key}: {value}".format(
                message=error['message'],
                key=key,
                value=value,
            ),
        ) for key, value in error["errors"].items()])

    def _on_missing_http_header(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def _on_invalid_http_header(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def _on_too_many_request(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def _on_method_not_allowed(self, error):
        return CLIError([dict(
            msg=error["message"],
        )])

    def get_unexpected_error(self):
        return CLIError([
            dict(
                msg="An unexpected error happened while running your command.",
                hint="Please run your command with --debug and send the output "
                     "to us via support@digicloud.ir",
            )
        ])



"""
    Allow to run digicloud as python module.

    Example:
        ``python3 -m digicloud account logout``
"""
import sys

from digicloud.main import main


if __name__ == '__main__':
    sys.exit(main())



"""
    Managers definition.
"""
import os
import pickle
import sys
from pathlib import Path
from collections import ChainMap

from requests import Session as Sess
from sentry_sdk import capture_message


class ConfigManager():
    """Configuration manager for all configs."""
    FILE_NAME = 'config'
    DEFAULTS = {
        'BASE_URL': os.environ.get('API_BASE_URL', "https://api.digicloud.ir"),
    }

    def __init__(self, app_logger):
        self._log = app_logger
        self._path = self._get_config_path()
        self._states = dict()
        self._load()

    def __getitem__(self, item):
        try:
            return ChainMap(self.DEFAULTS, self._states)[item]
        except KeyError:
            raise KeyError('No "{}" in configuration.'.format(item))

    def __setitem__(self, item, value):
        self._states[item] = value

    def __delitem__(self, item):
        if item in self._states:
            del self._states[item]

    def __contains__(self, item):
        return item in self._states

    def __call__(self):
        return self._states

    def _get_config_path(self):
        """Create config file directory.

        Handles platform dependent path creation for ``FILE_NAME ``,
        and return full path to config file.
        """
        path = {
            'darwin': '$HOME/Library/Application Support/digicloud',
            'win': '%LOCALAPPDATA%/digicloud',
            'linux': '$HOME/.config/digicloud',
        }[sys.platform]

        path = os.path.expandvars(path)

        if not os.path.isdir(path):
            os.makedirs(path, mode=0o755, exist_ok=True)

        full_path = os.path.join(path, self.FILE_NAME)
        Path(full_path).touch()

        return full_path

    def _load(self):
        with open(self._path, 'rb') as file_:
            if os.path.getsize(self._path) > 0:
                self._states = pickle.load(file_)

    def _dump(self):
        with open(self._path, 'wb') as file_:
            pickle.dump(self._states, file_, pickle.HIGHEST_PROTOCOL)

    def get(self, item, default=None):
        """return item from config or default."""
        return self._states.get(item, default)

    def save(self):
        """write configurations to file."""
        self._dump()


class Session(Sess):
    """Custom ``requests.Session`` object model."""

    def __init__(self, base_url, *args, **kwargs):
        self.base_url = base_url
        super(Session, self).__init__(*args, **kwargs)

    @property
    def user_agent(self):
        return self.headers.get("User-Agent")

    @user_agent.setter
    def user_agent(self, value):
        self.headers["User-Agent"] = value

    def resource(self, uri, method='GET', payload=None, params=None,
                 endpoint_version='/v1'):
        """Resource access helper."""
        url = f'{self.base_url}{endpoint_version}{uri}'
        response = self.request(method, url, params=params, json=payload)
        self.log_to_sentry(response)
        response.raise_for_status()
        if len(response.content):
            return response.json()
        return None

    def get(self, uri, params=None):
        return self.resource(uri, params=params)

    def post(self, uri, payload):
        return self.resource(uri, 'POST', payload)

    def put(self, uri, payload):
        return self.resource(uri, 'PUT', payload)

    def patch(self, uri, payload):
        return self.resource(uri, 'PATCH', payload)

    def delete(self, uri, payload=None):
        return self.resource(uri, 'DELETE', payload)

    def setup(self, config):
        self.base_url = config['BASE_URL']
        if config.get('AUTH_HEADERS'):
            auth_headers = config['AUTH_HEADERS']
            self.headers.update(auth_headers)
        else:
            config['AUTH_HEADERS'] = dict()

    def log_to_sentry(self, response):
        response_time = round(response.elapsed.total_seconds(), 1)
        if response_time > 25:
            capture_message('slow-response: url: {} time: {} seconds'.format(
                response.request.url, response_time
            ))



"""
    Authentication to digicloud API.
"""

from digicloud.commands.base import Command
from ..cli import parseractions
from ..error_handlers import CLIError


class Config(Command):
    """Managing Digicloud configurations"""

    def get_parser(self, prog_name):
        parser = super(Config, self).get_parser(prog_name)

        parser.add_argument(
            '--set',
            metavar='collect_error_info=<collect_error_info>',
            dest='configs',
            action=parseractions.MultiKeyValueAction,
            optional_keys=['collect_error_info', ],
            help='Set CLI configuration, for example --set collect_error_info=1/0'
        )

        return parser

    def take_action(self, parsed_args):
        if parsed_args.configs is None:
            self.show_configs()
            return
        if len(parsed_args.configs) > 1:
            raise CLIError([dict(msg="You only need to specify --set once")])
        app_config = self.app.config
        if not app_config.get('options'):
            app_config['options'] = {}
        options = app_config['options']
        for key, value in parsed_args.configs[0].items():
            self.app.console.print(
                "Setting [bold blue]%s[/bold blue] to [bold green]%s[/bold green] " % (
                    key, value
                )
            )
            options[key] = value
        app_config.save()

    def show_configs(self):
        app_config = self.app.config
        if not app_config.get('options'):
            app_config['options'] = {}
        options = app_config['options']
        if not options:
            self.app.console.print("You haven't set any configuration parameter")
            return
        for key, value in options.items():
            self.app.console.print(
                "[bold blue]%s[/bold blue]: [bold green]%s[/bold green]" % (
                    key, value
                )
            )



"""
    Digicloud Namespace Management.
"""

from .base import Lister, ShowOne, Command
from .. import schemas
from ..error_handlers import CLIError


def get_namespace(name_or_id, session):
    candidates = []
    for namespace in session.get('/namespaces'):
        if name_or_id == namespace['id'] or name_or_id == namespace['name']:
            candidates.append(namespace)
    if len(candidates) == 1:
        return candidates[0]
    if len(candidates) > 1:
        raise CLIError([dict(
            msg="You are a member of multiple namespaces named '{}'".format(name_or_id),
            hint="You should check list of your namespaces and repeat this command with"
                 " namespace id instead of name."
        )])
    raise CLIError([dict(
        msg="You're not a member of a namespace with this name or ID",
        hint="It might be because you leave "
             "the namespace (or someone removed you)."
             " you can see list of your namespaces by "
             "[blue bold]digicloud namespace list[/blue bold]"
    )])


class CreateNamespace(ShowOne):
    """Create Namespace."""
    schema = schemas.NamespaceDetails()

    def get_parser(self, prog_name):
        parser = super(CreateNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of namespace'
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help='Description of namespace'
        )
        return parser

    def get_data(self, parsed_args):
        return self.app.session.post('/namespaces', {
            'name': parsed_args.name,
            'description': parsed_args.description,
        })


class ListNamespace(Lister):
    """List namespaces"""
    schema = schemas.NamespaceList(many=True)

    def get_data(self, parsed_args):
        return self.app.session.get('/namespaces')


class ShowNamespace(ShowOne):
    """Show server details."""
    schema = schemas.NamespaceDetails()

    def get_parser(self, prog_name):
        parser = super(ShowNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )
        return parser

    def get_data(self, parsed_args):
        return get_namespace(parsed_args.namespace, self.app.session)


class ListNamespaceMember(Lister):
    """List namespace members and invitations."""
    schema = schemas.NamespaceMemberList(many=True)

    def get_parser(self, prog_name):
        parser = super(ListNamespaceMember, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )
        return parser

    def get_data(self, parsed_args):
        namespace = get_namespace(parsed_args.namespace, self.app.session)
        members = self.app.session.get("/namespaces/{id}/members".format(id=namespace['id']))
        invitations = self.app.session.get("/namespaces/{id}/invitations".format(id=namespace['id']))
        return members + invitations


class InviteMember(Command):
    """Add Member to Namespace."""

    def get_parser(self, prog_name):
        parser = super(InviteMember, self).get_parser(prog_name)
        parser.add_argument(
            '--email',
            metavar='<user_email>',
            required=True,
            help='E-mail Address to send the invitation'
        )

        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )
        return parser

    def take_action(self, parsed_args):
        namespace = get_namespace(parsed_args.namespace, self.app.session)
        uri = '/invitations'
        payload = {
            'email': parsed_args.email,
            'namespace': namespace['id']
        }
        self.app.session.post(uri, payload)
        self.app.console.print(
            "[green bold]An invitation has been sent to {}[/green bold]".format(
                parsed_args.email
            )
        )


class DeleteNamespaceMember(Command):
    """Delete Member from Namespace."""

    def get_parser(self, prog_name):
        parser = super(DeleteNamespaceMember, self).get_parser(prog_name)
        parser.add_argument(
            '--user-id',
            metavar='<user_id>',
            required=True,
            help='ID of user'
        )
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )

        return parser

    def take_action(self, parsed_args):
        namespace = get_namespace(parsed_args.namespace, self.app.session)
        uri = '/namespaces/{}/members/{}'.format(
            namespace["id"],
            parsed_args.user_id
        )
        self.app.session.delete(uri)


class LeaveNamespace(Command):
    """Leave a namespace """

    def get_parser(self, prog_name):
        parser = super(LeaveNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )

        return parser

    def take_action(self, parsed_args):
        current_user = self.app.config['USER']['id']
        namespace = get_namespace(parsed_args.namespace, self.app.session)
        uri = '/namespaces/%s/members/%s' % (namespace['id'], current_user)
        self.app.session.delete(uri)


class DeleteNamespace(Command):
    """Delete Namespace."""

    def get_parser(self, prog_name):
        parser = super(DeleteNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )
        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )

        return parser

    def take_action(self, parsed_args):
        self.app.console.print("Delete namespace feature is not currently available")


class SelectNamespace(ShowOne):
    """Select Namespace."""
    schema = schemas.NamespaceDetails()

    def get_parser(self, prog_name):
        parser = super(SelectNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<NAMESPACE ID>',
            help='Select namespace as default',
        )
        return parser

    def get_data(self, parsed_args):
        namespace = get_namespace(parsed_args.namespace, self.app.session)
        uid = self.app.config['USER']['id']
        url = '/users/%s' % uid
        payload = {'default_namespace': namespace['id']}
        self.app.session.patch(url, payload)
        self.app.config['AUTH_HEADERS']['Digicloud-Namespace'] = namespace['id']
        return namespace


class CurrentNamespace(ShowOne):
    """Display current namespace."""
    schema = schemas.NamespaceDetails()

    def get_parser(self, prog_name):
        parser = super(CurrentNamespace, self).get_parser(prog_name)

        return parser

    def get_data(self, parsed_args):
        if self.app.config.get('AUTH_HEADERS') is None:
            raise CLIError([
                dict(
                    msg="You need to login first",
                    hint="you can login by [blue bold]digicloud account login[/blue bold]"
                )
            ])

        namespace_id = self.app.config['AUTH_HEADERS']['Digicloud-Namespace']
        namespace = self.app.session.get('/namespaces/%s' % namespace_id)
        return namespace


class AcceptInvitation(ShowOne):
    """Accept an invitation to join a namespace"""
    schema = schemas.UserInvitation()

    def get_parser(self, prog_name):
        parser = super(AcceptInvitation, self).get_parser(prog_name)
        parser.add_argument(
            'invitation',
            metavar='<invitation>',
            help='Invitation ID to accept or reject'
        )

        return parser

    def get_data(self, parsed_args):
        uri = '/invitations/%s' % parsed_args.invitation
        payload = {
            'status': "accepted"
        }
        return self.app.session.patch(uri, payload)


class ListQuota(Lister):
    """List your namespace quota"""
    schema = schemas.NamespaceQuota(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<NAMESPACE>',
            help='Namespace name or id',
        )
        return parser

    def get_data(self, parsed_args):
        namespace = get_namespace(parsed_args.namespace, self.app.session)
        uri = '/quotas/{}'.format(namespace["id"])
        data = self.app.session.get(uri)
        return data


class ListUserInvitations(Lister):
    """Lists all user invitations"""
    schema = schemas.UserInvitation(many=True)

    def get_data(self, parsed_args):
        uri = '/invitations'
        return self.app.session.get(uri)
